#####################################################.
#            This file stores all the functions     #
#               used in conformer generation        #
#####################################################.

import math
import os
import sys
import subprocess
import time
from pathlib import Path

import numpy as np
import pandas as pd
from rdkit.Chem import AllChem as Chem

from rdkit.Chem import rdMolTransforms, PropertyMol, rdDistGeom, Lipinski

from aqme.filter import filters, ewin_filter, pre_E_filter, RMSD_and_E_filter
from aqme.tmbuild import template_embed
from aqme.cmin import rules_get_charge, substituted_mol
from aqme.fullmonte import (
	generating_conformations_fullmonte,
	minimize_rdkit_energy,
	realign_mol,
)
from aqme.utils import (
	Logger,
	set_metal_atomic_number,
	com_2_xyz_2_sdf,
	getDihedralMatches,
	load_from_yaml,
	creation_of_dup_csv_csearch,
	check_charge_smi,
	check_for_pieces,
	nci_ts_mol,
)
from aqme.crest import crest_opt

from aqme.argument_parser import set_options

SUPPORTED_INPUTS = [
	".smi",
	".sdf",
	".cdx",
	".csv",
	".com",
	".gjf",
	".mol",
	".mol2",
	".xyz",
	".txt",
	".yaml",
	".yml",
	".rtf",
]

# get pre-determined geometries for metal complexes
accepted_complex_types = ["squareplanar", "squarepyramidal", "linear", "trigonalplanar"]


def process_csearch(smi_, name_, dir_, varfile_, charge_default_, constraints_dist_, constraints_angle_, constraints_dihedral_):
	obj = csearch(smi_, name_, dir_, varfile_, charge_default_, constraints_dist_, constraints_angle_, constraints_dihedral_)
	total_data = obj.compute_confs()
	return total_data


class csearch:
	"""
	Representation of the neccesary information related with csearch.

	Parameters
	----------
	mol : RDKit mol object
        Mol object used in CSEARCH
    name : str
        Name of the mol object
    w_dir_initial : str
        Working directory
	charge_default : int
        Charge of the system
	varfile : str
        Parameter file containing the additional options for CMIN
	constraints_dist : list of lists
		If the conformational sampling is for a transition state, this option add distance constrains
		between sets of 2 atoms in the format [[A1,A2,DIST1],[A3,A4,DIST2]]. For example, constrain
		C atom 1 and O atom 2 to a distance of 2.0 A --> [[1,2,2.0]]
	constraints_angle : list of lists
		Same as constraints_dist but contraining angles. For example, constrain C atom 1, O atom 2,
		and C atom 3 to an angle of 120 degrees --> [[1,2,3,120]]
	constraints_dihedral : list of lists
		Same as constraints_dist but contraining dihedral angles. For example, constrain C atom 1, O atom 2,
		C atom 3, and O atom 4 to a dihedral angle of 120 degrees --> [[1,2,3,4,120]]
	kwargs : argument class
		Specify any arguments from the QCORR module
	"""

	def __init__(
		self,
		smi=None,
		name=None,
		w_dir_initial=os.getcwd(),
		varfile=None,
		charge_default=0,
		constraints_dist=[],
		constraints_angle=[],
		constraints_dihedral=[],
		**kwargs,
	):
		self.smi = smi
		self.name = name
		self.w_dir_initial = w_dir_initial
		self.charge_default = charge_default

		if "options" in kwargs:
			self.args = kwargs["options"]
		else:
			self.args = set_options(kwargs)
		self.args.varfile = varfile

		csearch_dir = Path(self.w_dir_initial) / "CSEARCH"
		dat_dir = csearch_dir / "dat_files"
		dat_dir.mkdir(parents=True, exist_ok=True)

		self.log = Logger(dat_dir / self.name, self.args.output_name)

		if varfile is not None:
			self.args, self.log = load_from_yaml(self.args, self.log)

		self.mol , self.args.constraints_dist, self.args.constraints_angle, self.args.constraints_dihedral  = smi_to_mol(self.smi, self.args, self.name, constraints_dist, constraints_angle, constraints_dihedral)

		self.args.charge_default = self.charge_default
		self.args.charge = rules_get_charge(self.mol, self.args)

		self.csearch_folder = Path(self.w_dir_initial).joinpath(f"CSEARCH/{self.args.CSEARCH}")
		self.csearch_folder.mkdir(exist_ok=True)
		self.csearch_file = []
		

	def compute_confs(self):

		"""
		function to start conf generation

		Parameters
		----------
		w_dir_initial : [type]    [description]
		mol : rdkit.Chem.Mol    [description]
		name : [type]    [description]
		i : [type]    [description]

		Returns
		-------
		pandas.Dataframe    total_data
		"""

		# Converts each line to a rdkit mol object
		if self.args.verbose:
			self.log.write(f"   -> Input Molecule {Chem.MolToSmiles(self.mol)}")

		if self.args.metal_complex:
			for _ in self.args.metal:
				self.args.metal_idx.append(None)
				self.args.complex_coord.append(None)
				self.args.metal_sym.append(None)

			(
				self.args.mol,
				self.args.metal_idx,
				self.args.complex_coord,
				self.args.metal_sym,
			) = substituted_mol(self.mol, self.args)

			# get pre-determined geometries for metal complexes
			accepted_complex_types = [
				"squareplanar",
				"squarepyramidal",
				"linear",
				"trigonalplanar",
			]
			if self.args.complex_type in accepted_complex_types:
				count_metals = 0
				for metal_idx_ind in self.args.metal_idx:
					if metal_idx_ind is not None:
						count_metals += 1
				if count_metals == 1:
					os.chdir(self.w_dir_initial)
					template_kwargs = dict()
					template_kwargs["complex_type"] = self.args.complex_type
					template_kwargs["metal_idx"] = self.args.metal_idx
					template_kwargs["maxsteps"] = self.args.opt_steps_RDKit
					template_kwargs["heavyonly"] = self.args.heavyonly
					template_kwargs["maxmatches"] = self.args.max_matches_RMSD
					items = template_embed(self.mol, self.name, self.log, **template_kwargs)

					total_data = creation_of_dup_csv_csearch(self.args.CSEARCH)

					for mol_obj, name_in, coord_map, alg_map, template in zip(*items):
						data = self.conformer_generation(
							mol_obj,
							name_in,
							self.args,
							self.log,
							coord_map,
							alg_map,
							template,
						)
						frames = [total_data, data]
						total_data = pd.concat(frames, sort=True)
				else:
					self.log.write(
						"x  Cannot use templates for complexes involving more than 1 metal or for organic molecueles."
					)
					total_data = None
			else:
				total_data = self.conformer_generation(
					self.mol, self.name, self.args, self.log
				)
		else:
			total_data = self.conformer_generation(
				self.mol, self.name, self.args, self.log
			)
		return total_data

	def conformer_generation(
		self, mol, name, args, log, coord_Map=None, alg_Map=None, mol_template=None
	):

		"""
		FUNCTION WORKING WITH MOL OBJECT TO CREATE CONFORMERS

		Parameters
		----------
		mol : rdkit.Chem.Mol    [description]
		name : [type]    [description]
		args : argparse.args    [description]
		log : [type]    [description]
		coord_Map : [type], optional    [description], by default None
		alg_Map : [type], optional    [description], by default None
		mol_template : [type], optional    [description], by default None

		Returns
		-------
		pd.Dataframe    dup_data
		"""
		dup_data = creation_of_dup_csv_csearch(args.CSEARCH)

		file = self.csearch_folder.joinpath(name + "_" + self.args.CSEARCH + self.args.output)
		self.csearch_file = self.csearch_file + [file.as_posix()]

		self.sdwriter = Chem.SDWriter(str(file))

		dup_data_idx = 0
		start_time = time.time()
		valid_structure = filters(mol, log, args.max_MolWt, args.verbose)
		if valid_structure:
			if args.verbose:
				log.write(f"\n   ----- {name} -----")
			try:
				# the conformational search for RDKit
				status, update_to_rdkit = self.summ_search(
					mol,
					name,
					args,
					log,
					dup_data,
					dup_data_idx,
					coord_Map,
					alg_Map,
					mol_template,
				)
				dup_data.at[dup_data_idx, "status"] = status
				dup_data.at[dup_data_idx, "update_to_rdkit"] = update_to_rdkit
			except (KeyboardInterrupt, SystemExit):  # RAUL: This try-except is useless.
				raise
		else:
			log.write("\nx  ERROR: The structure is not valid")

		if args.time:
			n_seconds = round(time.time() - start_time, 2)
			log.write(f"\n Execution time CSEARCH: {n_seconds} seconds")
			dup_data.at[dup_data_idx, "CSEARCH time (seconds)"] = n_seconds
		return dup_data

	def summ_search(
		self,
		mol,
		name,
		args,
		log,
		dup_data,
		dup_data_idx,
		coord_Map=None,
		alg_Map=None,
		mol_template=None,
	):

		"""
		EMBEDS, OPTIMIZES AND FILTERS RDKIT CONFORMERS

		Parameters
		----------
		mol : [type]    [description]
		name : [type]    [description]
		args : [type]    [description]
		log : [type]    [description]
		dup_data : [type]    [description]
		dup_data_idx : [type]    [description]
		coord_Map : [type], optional    [description], by default None
		alg_Map : [type], optional    [description], by default None
		mol_template : [type], optional    [description], by default None

		Returns
		-------
		tuple    status, update_to_rdkit
		"""

		# writes sdf for the first RDKit conformer generation
		status, rotmatches, update_to_rdkit = self.rdkit_to_sdf(
			mol,
			name,
			args,
			log,
			dup_data,
			dup_data_idx,
			self.sdwriter,
			coord_Map,
			alg_Map,
			mol_template

		)
		# reads the initial SDF files from RDKit and uses dihedral scan if selected
		if status != -1 or status != 0:
			# getting the energy and mols after rotations
			if args.CSEARCH == "summ" and len(rotmatches) != 0:
				status = self.dihedral_filter_and_sdf(
					name,
					args,
					log,
					dup_data,
					dup_data_idx,
					coord_Map,
					alg_Map,
					mol_template,
				)

				# removes the rdkit file
				os.remove(name + "_" + "rdkit" + args.output)

		return status, update_to_rdkit

	def dihedral_filter_and_sdf(
		self, name, args, log, dup_data, dup_data_idx, coord_Map, alg_Map, mol_template
	):
		"""
		filtering after dihedral scan to sdf

		Parameters
		----------
		name : [type]    [description]
		args : [type]    [description]
		log : [type]    [description]
		dup_data : [type]    [description]
		dup_data_idx : [type]    [description]
		coord_Map : [type]    [description]
		alg_Map : [type]    [description]
		mol_template : [type]    [description]

		Returns
		-------
		int    status (job I guess?)
		"""
		rotated_energy = []

		rdmols = Chem.SDMolSupplier(name + "_" + "rdkit" + args.output, removeHs=False)

		if rdmols is None:
			log.write("Could not open " + name + args.output)
			sys.exit(-1)

		for i, rd_mol_i in enumerate(rdmols):
			if coord_Map is None and alg_Map is None and mol_template is None:
				energy = minimize_rdkit_energy(
					rd_mol_i, -1, log, args.ff, args.opt_steps_RDKit
				)
			else:
				rd_mol_i, energy = realign_mol(
					rd_mol_i, -1, coord_Map, alg_Map, mol_template, args.opt_steps_RDKit
				)
			rotated_energy.append(energy)

		rotated_cids = list(range(len(rdmols)))
		sorted_rotated_cids = sorted(rotated_cids, key=lambda cid: rotated_energy[cid])

		# filter based on energy window ewin_csearch
		sortedcids_rotated = ewin_filter(
			sorted_rotated_cids,
			rotated_energy,
			args,
			dup_data,
			dup_data_idx,
			log,
			"summ",
			args.ewin_csearch,
		)
		# pre-filter based on energy only
		selectedcids_initial_rotated = pre_E_filter(
			sortedcids_rotated,
			rotated_energy,
			dup_data,
			dup_data_idx,
			log,
			"summ",
			args.initial_energy_threshold,
			args.verbose,
		)
		# filter based on energy and RMSD
		selectedcids_rotated = RMSD_and_E_filter(
			rdmols,
			selectedcids_initial_rotated,
			rotated_energy,
			args,
			dup_data,
			dup_data_idx,
			log,
			"summ",
		)

		sdwriter_rd = Chem.SDWriter(name + "_" + "summ" + args.output)
		for i, cid in enumerate(selectedcids_rotated):
			mol_rd = Chem.RWMol(rdmols[cid])
			mol_rd.SetProp("_Name", rdmols[cid].GetProp("_Name") + " " + str(i))
			mol_rd.SetProp("Energy", str(rotated_energy[cid]))
			if args.metal_complex:
				set_metal_atomic_number(mol_rd, args.metal_idx, args.metal_sym)
			sdwriter_rd.write(mol_rd)
		sdwriter_rd.close()
		status = 1
		return status

	def clean_args(
		self, args, ori_ff, mol, ori_charge
	):  # RAUL: I hope this function does not survive the clean-ups - Shree : not needed anymore (just leaving it in here for now)
		"""
		returns the arguments to their original value after each calculation

		Parameters
		----------
		args : argparse.args    [description]
		ori_ff : [type]    original forcefield
		mol : rdkit.Chem.Mol    [description]
		ori_charge : [type]    original charge
		"""
		for atom in mol.GetAtoms():
			if atom.GetSymbol() in args.metal:
				args.metal_complex = True
				break
		else:
			args.metal_complex = False
		args.ff = ori_ff
		args.charge_default = ori_charge
		args.metal_idx = []
		args.complex_coord = []
		args.metal_sym = []

	def auto_sampling(self, mult_factor, mol, args, log):
		"""
		DETECTS INITIAL NUMBER OF SAMPLES AUTOMATICALLY

		Parameters
		----------
		mult_factor : [type]    [description]
		mol : [type]    [description]
		args : [type]    [description]
		log : [type]    [description]

		Returns
		-------
		int    auto_samples
		"""
		if args.metal_complex:
			if len(args.metal_idx) > 0:
				mult_factor = (
					mult_factor * 3 * len(args.metal_idx)
				)  # this accounts for possible trans/cis isomers in metal complexes
		auto_samples = 0
		auto_samples += 3 * (Lipinski.NumRotatableBonds(mol))  # x3, for C3 rotations
		auto_samples += 3 * (Lipinski.NHOHCount(mol))  # x3, for OH/NH rotations
		auto_samples += 3 * (
			Lipinski.NumSaturatedRings(mol)
		)  # x3, for boat/chair/envelope confs
		if auto_samples == 0:
			auto_samples = mult_factor
		else:
			auto_samples = mult_factor * auto_samples
		return auto_samples

	def genConformer_r(
		self,
		mol,
		conf,
		i,
		matches,
		degree,
		sdwriter,
		args,
		name,
		log,
		update_to_rdkit,
		coord_Map,
		alg_Map,
		mol_template,
	):
		"""
		IF NOT USING DIHEDRALS, THIS REPLACES I BACK TO THE METAL WHEN METAL = TRUE
		AND WRITES THE RDKIT SDF FILES. WITH DIHEDRALS, IT OPTIMIZES THE ROTAMERS

		Parameters
		----------
		mol : rdkit.Chem.Mol    [description]
		conf : [type]    [description]
		i : [type]    [description]
		matches : [type]    [description]
		degree : [type]    [description]
		sdwriter : [type]    [description]
		args : [type]    [description]
		name : [type]    [description]
		log : [type]    [description]
		update_to_rdkit : [type]    [description]
		coord_Map : [type]    [description]
		alg_Map : [type]    [description]
		mol_template : [type]    [description]

		Returns
		-------
		int    total number of conformers generated
		"""
		if i >= len(matches):  # base case, torsions should be set in conf
			# setting the metal back instead of I
			if args.metal_complex and (args.CSEARCH == "rdkit" or update_to_rdkit):
				if coord_Map is None and alg_Map is None and mol_template is None:
					energy = minimize_rdkit_energy(
						mol, conf, log, args.ff, args.opt_steps_RDKit
					)
				else:
					mol, energy = realign_mol(
						mol,
						conf,
						coord_Map,
						alg_Map,
						mol_template,
						args.opt_steps_RDKit,
					)
				mol.SetProp("Energy", str(energy))
				set_metal_atomic_number(mol, args.metal_idx, args.metal_sym)
			try:
				sdwriter.write(mol,conf)
			except:
				pass
			return 1
		else:
			total = 0
			deg = 0
			while deg < 360.0:
				rad = math.pi * deg / 180.0
				rdMolTransforms.SetDihedralRad(
					mol.GetConformer(conf), *matches[i], value=rad
				)
				mol.SetProp("_Name", name)
				total += self.genConformer_r(
					mol,
					conf,
					i + 1,
					matches,
					degree,
					sdwriter,
					args,
					name,
					log,
					update_to_rdkit,
					coord_Map,
					alg_Map,
					mol_template,
				)
				deg += degree
			return total

	def embed_conf(
		self, mol, initial_confs, args, log, coord_Map, alg_Map, mol_template
	):
		"""
		function to embed conformers

		Parameters
		----------
		mol : [type]    [description]
		initial_confs : [type]    [description]
		args : [type]    [description]
		log : [type]    [description]
		coord_Map : [type]    [description]
		alg_Map : [type]    [description]
		mol_template : [type]    [description]

		Returns
		-------
		list    cids
		"""
		is_sdf_mol_or_mol2 = os.path.splitext(args.input)[1] in [
			".sdf",
			".mol",
			".mol2",
		]

		if is_sdf_mol_or_mol2:
			Chem.AssignStereochemistryFrom3D(mol)

		embed_kwargs = dict()
		embed_kwargs["ignoreSmoothingFailures"] = True
		embed_kwargs["randomSeed"] = args.seed
		embed_kwargs["numThreads"] = 0

		if (coord_Map, alg_Map, mol_template) != (None, None, None):
			embed_kwargs["coordMap"] = coord_Map
		cids = rdDistGeom.EmbedMultipleConfs(mol, initial_confs, **embed_kwargs)

		if len(cids) <= 1 and initial_confs != 1:
			log.write(
				"o  Normal RDKit embeding process failed, trying to "
				"generate conformers with random coordinates "
				f"(with {str(initial_confs)} possibilities)"
			)
			embed_kwargs["useRandomCoords"] = True
			embed_kwargs["boxSizeMult"] = 10.0
			embed_kwargs["numZeroFail"] = 1000
			embed_kwargs["numThreads"] = 1
			cids = rdDistGeom.EmbedMultipleConfs(mol, initial_confs, **embed_kwargs)

		if is_sdf_mol_or_mol2:
			# preserving AssignStereochemistryFrom3D
			for cid in cids:
				Chem.AssignAtomChiralTagsFromStructure(mol, confId=cid)

		return cids

	def min_and_E_calc(self, mol, cids, args, log, coord_Map, alg_Map, mol_template):
		"""
		minimization and E calculation with RDKit after embeding

		Parameters
		----------
		mol : [type] [description]
		cids : [type][description]
		args : [type][description]
		log : [type][description]
		coord_Map : [type][description]
		alg_Map : [type]    [description]
		mol_template : [type]    [description]

		Returns
		-------
		outmols,cenergy    outmols is gonna be a list containing "initial_confs" mol objects    with "initial_confs" conformers. We do this to SetProp    (Name and Energy) to the different conformers    and log.write in the SDF file. At the end, since all the mol    objects has the same conformers, but the energies are different,    we can log.write conformers to SDF files with the energies of the    parent mol objects. We measured the computing time and it's the    same as using only 1 parent mol object with 10 conformers, but we    couldn'temp SetProp correctly.
		"""

		cenergy, outmols = [], []
		# bar = IncrementalBar('o  Minimizing', max = len(cids))
		for _, conf in enumerate(cids):
			if coord_Map is None and alg_Map is None and mol_template is None:
				energy = minimize_rdkit_energy(
					mol, conf, log, args.ff, args.opt_steps_RDKit
				)
			else:  # id template realign before doing calculations
				mol, energy = realign_mol(
					mol, conf, coord_Map, alg_Map, mol_template, args.opt_steps_RDKit
				)
			cenergy.append(energy)
			pmol = PropertyMol.PropertyMol(mol)
			outmols.append(pmol)
			# bar.next()
		# bar.finish()
		return outmols, cenergy

	def min_after_embed(
		self,
		mol,
		cids,
		name,
		initial_confs,
		rotmatches,
		dup_data,
		dup_data_idx,
		sdwriter,
		args,
		log,
		update_to_rdkit,
		coord_Map,
		alg_Map,
		mol_template,
	):
		"""
		minimizes, gets the energy and filters RDKit conformers after embeding

		Parameters
		----------
		mol : [type]    [description]
		cids : [type]    [description]
		name : [type]    [description]
		initial_confs : [type]    [description]
		rotmatches : [type]    [description]
		dup_data : [type]    [description]
		dup_data_idx : [type]    [description]
		sdwriter : [type]    [description]
		args : [type]    [description]
		log : [type]    [description]
		update_to_rdkit : [type]    [description]
		coord_Map : [type]    [description]
		alg_Map : [type]    [description]
		mol_template : [type]    [description]

		Returns
		-------
		int    status
		"""

		# gets optimized mol objects and energies
		outmols, cenergy = self.min_and_E_calc(
			mol, cids, args, log, coord_Map, alg_Map, mol_template
		)

		# writing charges after RDKit
		if (
			os.path.splitext(args.input)[1] == ".cdx"
			or os.path.splitext(args.input)[1] == ".smi"
			or os.path.splitext(args.input)[1] == ".csv"
		):
			dup_data.at[dup_data_idx, "Overall charge"] = np.sum(args.charge)
		else:
			dup_data.at[dup_data_idx, "Overall charge"] = args.charge_default

		for i, cid in enumerate(cids):
			outmols[cid].SetProp("_Name", name + " " + str(i + 1))
			outmols[cid].SetProp("Energy", str(cenergy[cid]))
			outmols[cid].SetProp(
				"Real charge", str(dup_data.at[dup_data_idx, "Overall charge"])
			)

		# sorts the energies
		cids = list(range(len(outmols)))
		sorted_all_cids = sorted(cids, key=lambda cid: cenergy[cid])

		log.write("\n\no  Applying filters to intial conformers")

		# filter based on energy window ewin_csearch
		sortedcids_rdkit = ewin_filter(
			sorted_all_cids,
			cenergy,
			args,
			dup_data,
			dup_data_idx,
			log,
			"rdkit",
			args.ewin_csearch,
		)

		# pre-filter based on energy only
		selectedcids_initial_rdkit = pre_E_filter(
			sortedcids_rdkit,
			cenergy,
			dup_data,
			dup_data_idx,
			log,
			"rdkit",
			args.initial_energy_threshold,
			args.verbose,
		)

		# filter based on energy and RMSD
		selectedcids_rdkit = RMSD_and_E_filter(
			outmols,
			selectedcids_initial_rdkit,
			cenergy,
			args,
			dup_data,
			dup_data_idx,
			log,
			"rdkit",
		)

		if args.CSEARCH == "summ" or args.CSEARCH == "rdkit":
			# now exhaustively drive torsions of selected conformers
			n_confs = int(
				len(selectedcids_rdkit) * (360 / args.degree) ** len(rotmatches)
			)
			if args.verbose and len(rotmatches) != 0:
				log.write(
					"\n\no  Systematic generation of " + str(n_confs) + " confomers"
				)
				# bar = IncrementalBar('o  Generating conformations based on dihedral rotation', max = len(selectedcids_rdkit))
			# else:
			#     bar = IncrementalBar('o  Writing unique conformers into an sdf file', max = len(selectedcids_rdkit))

			total = 0
			for conf in selectedcids_rdkit:
				if args.CSEARCH == "summ" and not update_to_rdkit:
					sdwriter.write(outmols[conf], conf)
					for m in rotmatches:
						rdMolTransforms.SetDihedralDeg(
							outmols[conf].GetConformer(conf), *m, 180.0
						)
				total += self.genConformer_r(
					outmols[conf],
					conf,
					0,
					rotmatches,
					args.degree,
					sdwriter,
					args,
					outmols[conf].GetProp("_Name"),
					log,
					update_to_rdkit,
					coord_Map,
					alg_Map,
					mol_template,
				)
				# bar.next()
			# bar.finish()
			if args.verbose and len(rotmatches) != 0:
				log.write("o  %d total conformations generated" % total)
			status = 1

		if args.CSEARCH == "summ":
			dup_data.at[dup_data_idx, "summ-conformers"] = total

		if args.CSEARCH == "fullmonte":
			status = generating_conformations_fullmonte(
				name,
				args,
				rotmatches,
				log,
				selectedcids_rdkit,
				outmols,
				sdwriter,
				dup_data,
				dup_data_idx,
				coord_Map,
				alg_Map,
				mol_template,
			)
			# removes the rdkit file
			os.remove(name + "_" + "rdkit" + args.output)

		return status

	def rdkit_to_sdf(
		self,
		mol,
		name,
		args,
		log,
		dup_data,
		dup_data_idx,
		sdwriter,
		coord_Map,
		alg_Map,
		mol_template
	):

		"""
		conversion from rdkit to sdf

		Parameters
		----------
		mol : [type]    [description]
		name : [type]    [description]
		args : [type]    [description]
		log : [type]    [description]
		dup_data : [type]    [description]
		dup_data_idx : [type]    [description]
		coord_Map : [type]    [description]
		alg_Map : [type]    [description]
		mol_template : [type]    [description]

		Returns
		-------
		tuple    status,rotmatches,update_to_rdkit
		"""
		Chem.SanitizeMol(mol)

		mol = Chem.AddHs(mol)

		mol.SetProp("_Name", name)

		# detects and applies auto-detection of initial number of conformers
		if args.sample == "auto":
			initial_confs = int(self.auto_sampling(args.auto_sample, mol, args, log))
		else:
			initial_confs = int(args.sample)

		dup_data.at[dup_data_idx, "Molecule"] = name
		update_to_rdkit = False

		rotmatches = getDihedralMatches(mol, args.heavyonly, log)

		if len(rotmatches) > args.max_torsions:
			log.write(
				"x  Too many torsions (%d). Skipping %s"
				% (len(rotmatches), (name + args.output))
			)
			status = -1
		elif args.CSEARCH == "summ" and len(rotmatches) == 0:
			update_to_rdkit = True
			log.write(
				"\nx  No rotatable dihedral found. Updating to CSEARCH to RDKit, writing to SUMM SDF"
			)
		elif args.CSEARCH == "fullmonte" and len(rotmatches) == 0:
			update_to_rdkit = True
			log.write(
				"\nx  No rotatable dihedral found. Updating to CSEARCH to RDKit, writing to FULLMONTE SDF"
			)

		# csearch_folder = Path(self.w_dir_initial).joinpath(f"CSEARCH/{args.CSEARCH}")
		# csearch_folder.mkdir(exist_ok=True)
		# csearch_file = csearch_folder.joinpath(name + "_" + args.CSEARCH + args.output)
		# sdwriter = Chem.SDWriter(str(csearch_file))

		# if update_to_rdkit and args.CSEARCH == "summ":
		#     sdwriter = Chem.SDWriter(name + "_" + "summ" + args.output)
		# elif update_to_rdkit and args.CSEARCH == "fullmonte":
		#     sdwriter = Chem.SDWriter(name + "_" + "fullmonte" + args.output)
		# elif args.CSEARCH == "fullmonte":
		#     sdwriter = Chem.SDWriter(name + "_" + "fullmonte" + args.output)
		# elif args.CSEARCH == "crest":
		#     sdwriter = Chem.SDWriter(name + "_" + "crest" + args.output)
		# else:
		#     sdwriter = Chem.SDWriter(name + "_" + "rdkit" + args.output)

		if args.CSEARCH != "crest":
			dup_data.at[dup_data_idx, "RDKit-Initial-samples"] = initial_confs
			if args.CSEARCH == "rdkit":
				rotmatches = []
			cids = self.embed_conf(
				mol, initial_confs, args, log, coord_Map, alg_Map, mol_template
			)

			# energy minimize all to get more realistic results
			# identify the atoms and decide Force Field
			for atom in mol.GetAtoms():
				if (
					atom.GetAtomicNum() > 36
				):  # up to Kr for MMFF, if not the code will use UFF
					log.write(
						"x  "
						+ args.ff
						+ " is not compatible with the molecule, changing to UFF"
					)
					args.ff = "UFF"
			if args.verbose:
				log.write(
					"o  Optimizing "
					+ str(len(cids))
					+ " initial conformers with "
					+ args.ff
				)
				if args.CSEARCH == "summ":
					log.write(
						"o  Found " + str(len(rotmatches)) + " rotatable torsions"
					)
				elif args.CSEARCH == "fullmonte":
					log.write(
						"o  Found " + str(len(rotmatches)) + " rotatable torsions"
					)
				else:
					log.write("o  Systematic torsion rotation is set to OFF")

			status = self.min_after_embed(
				mol,
				cids,
				name,
				initial_confs,
				rotmatches,
				dup_data,
				dup_data_idx,
				sdwriter,
				args,
				log,
				update_to_rdkit,
				coord_Map,
				alg_Map,
				mol_template,
			)
		else:
			args.charge = rules_get_charge(mol, args)
			dup_data.at[dup_data_idx, "Overall charge"] = np.sum(args.charge)
			status = crest_opt(mol, name, dup_data, dup_data_idx, sdwriter, args, log)

		sdwriter.close()

		return status, rotmatches, update_to_rdkit

def smi_to_mol(smi,args,name, constraints_dist, constraints_angle, constraints_dihedral):
	smi = check_for_pieces(smi)
	if len(smi)>1:
		mol, constraints_dist, constraints_angle, constraints_dihedral = nci_ts_mol(smi, args, constraints_dist, constraints_angle, constraints_dihedral, name)
	else:
		params = Chem.SmilesParserParams()
		params.removeHs = False
		mol = Chem.MolFromSmiles(smi[0],params)
	return mol, constraints_dist, constraints_angle, constraints_dihedral

# MAIN FUNCTION

# main function to generate conformers

def prepare_smiles_files(args, w_dir_initial):
	with open(args.input) as smifile:
		lines = [line for line in smifile if line.strip()]
	job_inputs = []
	for i, line in enumerate(lines):
		smi, name, args, constraints_dist, constraints_angle, constraints_dihedral = prepare_smiles_from_line(line, i, args)
		obj = smi, name, w_dir_initial, args.varfile, args.charge_default, constraints_dist, constraints_angle, constraints_dihedral
		job_inputs.append(obj)
	return job_inputs


def prepare_smiles_from_line(line, i, args):
	toks = line.split()
	# editing part
	smiles = toks[0]
	smi = check_for_pieces(smiles)
	if args.prefix == "None":  # I assume no AttributeError
		name = "".join(toks[1])  # I assume no AttributeError
	else:  # I assume no AttributeError
		name = f"{args.prefix}_{i}_{''.join(toks[1])}"  # I assume no AttributeError
	if len(smi)>1:
		constraints_angle, constraints_dist, constraints_dihedral = None, None, None
		if len(toks) > 2:
			constraints_dist = toks[2]
			constraints_dist = constraints_dist.split('/')
			for i,c in enumerate(constraints_dist):
				constraints_dist[i] = c.split('-')
			if len(toks) > 3:
				constraints_angle = toks[3]
				constraints_angle = constraints_angle.split('/')
				for i,c in enumerate(constraints_angle):
					constraints_angle[i] = c.split('-')
				if len(toks) > 4:
					constraints_dihedral = toks[4]
					constraints_dihedral = constraints_dihedral.split('/')
					for i,c in enumerate(constraints_dihedral):
						constraints_dihedral[i] = c.split('-')
	if args.charge_default == "auto":  # I assume no AttributeError
		if not args.metal_complex:  # I assume no AttributeError
			args.charge_default = check_charge_smi(smi, args.ts_complex)  # I assume no AttributeError
	return smiles, name, args, constraints_dist, constraints_angle, constraints_dihedral

def prepare_csv_files(args, w_dir_initial):
	csv_smiles = pd.read_csv(args.input)
	job_inputs = []
	for i in range(len(csv_smiles)):
		obj = generate_mol_from_csv(args, w_dir_initial, csv_smiles, i)
		job_inputs.append(obj)
	return job_inputs


def generate_mol_from_csv(args, w_dir_initial, csv_smiles, index):
	# assigning names and smi i  each loop
	smiles = csv_smiles.loc[index, "SMILES"]
	# pruned_smi = check_for_pieces(smi)
	# mol = Chem.MolFromSmiles(pruned_smi)

	smi = check_for_pieces(smiles)
	if args.prefix == "None":
		name = csv_smiles.loc[index, "code_name"]
	else:
		name = "comp_" + str(index) + "_" + csv_smiles.loc[index, "code_name"]
	constraints_angle, constraints_dist, constraints_dihedral = None, None, None
	if 'constraints_dist' in csv_smiles.columns:
		constraints_dist = csv_smiles.loc[index, "constraints_dist"]
		constraints_dist = constraints_dist.split('/')
		for i,c in enumerate(constraints_dist):
			constraints_dist[i] = c.split('-')
	if 'constraints_angle' in csv_smiles.columns:
		constraints_angle = csv_smiles.loc[index, "constraints_angle"]
		constraints_angle = constraints_angle.split('/')
		for i,c in enumerate(constraints_angle):
			constraints_angle[i] = c.split('-')
	if 'constraints_dihedral' in csv_smiles.columns:
		constraints_dihedral = csv_smiles.loc[index, "constraints_dihedral"]
		constraints_dihedral = constraints_dihedral.split('/')
		for i,c in enumerate(constraints_dihedral):
			constraints_dihedral[i] = c.split('-')
	if args.charge_default == "auto":
		if not args.metal_complex:
			args.charge_default = check_charge_smi(smi,args.ts_complex)
	obj = smiles, name, w_dir_initial, args.varfile, args.charge_default, constraints_dist, constraints_angle, constraints_dihedral
	return obj


def prepare_cdx_files(args, w_dir_initial):
	# converting to smiles from chemdraw
	molecules = generate_mol_from_cdx(args)
	job_inputs = []
	for i, (smi, mol) in enumerate(molecules):
		name = f"{args.input.split('.')[0]}_{str(i)}"
		if args.charge_default == "auto":
			if not args.metal_complex:
				args.charge_default = check_charge_smi(smi,args.ts_complex)
		constraints = []
		obj = mol, name, w_dir_initial, args.varfile, args.charge_default, constraints
		job_inputs.append(obj)
	return job_inputs


def generate_mol_from_cdx(args):
	cmd_cdx = ["obabel", "-icdx", args.input, "-osmi", "-Ocdx.smi"]
	subprocess.call(cmd_cdx)
	with open("cdx.smi", "r") as smifile:
		smi_lines = [line for line in smifile]
	os.remove("cdx.smi")
	molecules = []
	for smi in smi_lines:
		pruned_smi = check_for_pieces(smi)
		molecule = Chem.MolFromSmiles(pruned_smi)
		molecules.append((pruned_smi, molecule))
	return molecules


def prepare_gaussian_files(args, w_dir_initial):
	job_inputs = []
	charge_com = com_2_xyz_2_sdf(args.input, args.default_charge)
	name = os.path.splitext(args.input)[0]
	sdffile = f"{name}.sdf"
	suppl, _, _ = mol_from_sdf_or_mol_or_mol2(sdffile)

	for i, mol in enumerate(suppl):
		if args.charge_default == "auto":
			args.charge_default = charge_com
		constraints = []
		obj = mol, name, w_dir_initial, args.varfile, args.charge_default, constraints
		job_inputs.append(obj)
	return job_inputs


def prepare_xyz_files(args, w_dir_initial):
	job_inputs = []
	name = os.path.splitext(args.input)[0]
	sdffile = f"{name}.sdf"
	suppl, _, charge_com_list = mol_from_sdf_or_mol_or_mol2(sdffile)
	charge_com = charge_com_list[0]

	for i, mol in enumerate(suppl):
		if args.charge_default == "auto":
			args.charge_default = charge_com
		constraints = []
		obj = mol, name, w_dir_initial, args.varfile, args.charge_default, constraints
		job_inputs.append(obj)
	return job_inputs


def prepare_sdf_file(charge_sdf, w_dir_initial, mol, name, args, i):
	if args.charge_default == "auto":
		args.charge_default = charge_sdf
	constraints = []
	obj = mol, name, w_dir_initial, args.varfile, args.charge_default, constraints
	return obj


def prepare_sdf_files(args, w_dir_initial):
	suppl, IDs, charges = mol_from_sdf_or_mol_or_mol2(args.input)
	job_inputs = []
	for i, (mol, name, charge_sdf) in enumerate(zip(suppl, IDs, charges)):
		obj = prepare_sdf_file(charge_sdf, w_dir_initial, mol, name, args, i)
		job_inputs.append(obj)
	return job_inputs


def prepare_mol_file(suppl, name, charge, args, w_dir_initial):
	if args.charge_default == "auto":
		args.charge_default = charge
	mol = suppl
	constraints= []
	obj = mol, name, w_dir_initial, args.varfile, args.charge_default, constraints
	return obj


def prepare_mol_files(
	args, w_dir_initial
):  # The extra variables are for API Consistency.
	suppl, IDs, charges = mol_from_sdf_or_mol_or_mol2(args.input)
	job_inputs = []
	obj = prepare_mol_file(suppl, IDs[0], charges[0], args, w_dir_initial)
	job_inputs.append(obj)
	return job_inputs



"""
This module contains some classes and functions that are used from other modules
"""
import sys
import subprocess
import numpy as np
import glob
from pathlib import Path
import json
from rdkit.Chem.rdMolAlign import GetBestRMS
from rdkit.Chem.rdmolops import RemoveHs
from openbabel import pybel
from rdkit.Chem import AllChem
import os
from rdkit import Geometry
from rdkit.Chem import Draw
from rdkit.Chem import rdmolfiles
import yaml
import pandas as pd
from rdkit.Chem import AllChem as Chem


def periodic_table():
	items = """X
			H                                                                                                  He
			Li Be  B                                                                             C   N   O   F  Ne
			Na Mg Al                                                                            Si   P   S  Cl  Ar
			K Ca Sc                                           Ti  V Cr Mn Fe Co Ni Cu  Zn  Ga  Ge  As  Se  Br  Kr
			Rb Sr  Y                                           Zr Nb Mo Tc Ru Rh Pd Ag  Cd  In  Sn  Sb  Te   I  Xe
			Cs Ba La Ce Pr Nd Pm Sm Eu Gd Tb Dy Ho Er Tm Yb Lu Hf Ta  W Re Os Ir Pt Au  Hg  Tl  Pb  Bi  Po  At  Rn
			Fr Ra Ac Th Pa  U Np Pu Am Cm Bk Cf Es Fm Md No Lr Rf Db Sg Bh Hs Mt Ds Rg Uub Uut Uuq Uup Uuh Uus Uuo
			"""
	periodic_table = items.replace("\n", " ").strip().split()
	periodic_table[0] = ""
	return periodic_table


## Bondi VDW radii in Angstrom
def bondi():
	bondi = {
		"Bq": 0.00,
		"H": 1.09,
		"He": 1.40,
		"Li": 1.81,
		"Be": 1.53,
		"B": 1.92,
		"C": 1.70,
		"N": 1.55,
		"O": 1.52,
		"F": 1.47,
		"Ne": 1.54,
		"Na": 2.27,
		"Mg": 1.73,
		"Al": 1.84,
		"Si": 2.10,
		"P": 1.80,
		"S": 1.80,
		"Cl": 1.75,
		"Ar": 1.88,
		"K": 2.75,
		"Ca": 2.31,
		"Ni": 1.63,
		"Cu": 1.40,
		"Zn": 1.39,
		"Ga": 1.87,
		"Ge": 2.11,
		"As": 1.85,
		"Se": 1.90,
		"Br": 1.83,
		"Kr": 2.02,
		"Rb": 3.03,
		"Sr": 2.49,
		"Pd": 1.63,
		"Ag": 1.72,
		"Cd": 1.58,
		"In": 1.93,
		"Sn": 2.17,
		"Sb": 2.06,
		"Te": 2.06,
		"I": 1.98,
		"Xe": 2.16,
		"Cs": 3.43,
		"Ba": 2.68,
		"Pt": 1.72,
		"Au": 1.66,
		"Hg": 1.55,
		"Tl": 1.96,
		"Pb": 2.02,
		"Bi": 2.07,
		"Po": 1.97,
		"At": 2.02,
		"Rn": 2.20,
		"Fr": 3.48,
		"Ra": 2.83,
		"U": 1.86,
	}
	return bondi


## covalent radii in Angstrom (taken from Pyykko and Atsumi, Chem. Eur. J. 15, 2009, 188-197)
def rcov():
	rcov = {
		"H": 0.32,
		"He": 0.46,
		"Li": 1.33,
		"Be": 1.02,
		"B": 0.85,
		"C": 0.75,
		"N": 0.71,
		"O": 0.63,
		"F": 0.64,
		"Ne": 0.67,
		"Na": 1.55,
		"Mg": 1.39,
		"Al": 1.26,
		"Si": 1.16,
		"P": 1.11,
		"S": 1.03,
		"Cl": 0.99,
		"Ar": 0.96,
		"K": 1.96,
		"Ca": 1.71,
		"Sc": 1.48,
		"Ti": 1.36,
		"V": 1.34,
		"Cr": 1.22,
		"Mn": 1.19,
		"Fe": 1.16,
		"Co": 1.11,
		"Ni": 1.10,
		"Zn": 1.18,
		"Ga": 1.24,
		"Ge": 1.21,
		"As": 1.21,
		"Se": 1.16,
		"Br": 1.14,
		"Kr": 1.17,
		"Rb": 2.10,
		"Sr": 1.85,
		"Y": 1.63,
		"Zr": 1.54,
		"Nb": 1.47,
		"Mo": 1.38,
		"Tc": 1.28,
		"Ru": 1.25,
		"Rh": 1.25,
		"Pd": 1.20,
		"Ag": 1.28,
		"Cd": 1.36,
		"In": 1.42,
		"Sn": 1.40,
		"Sb": 1.40,
		"Te": 1.36,
		"I": 1.33,
		"Xe": 1.31,
	}
	return rcov


# load paramters from yaml file
def load_from_yaml(args_, log):
	"""
	Loads the parameters for the calculation from a yaml if specified. Otherwise
	does nothing.

	Parameters
	----------
	args : argparse.args
		Dataclass
	log : Logger
		Where to log the program progress
	"""
	# Variables will be updated from YAML file
	try:
		if args_.varfile is not None:
			if os.path.exists(args_.varfile):
				if os.path.splitext(args_.varfile)[1] == ".yaml":
					log.write("\no  Importing AQME parameters from " + args_.varfile)
					with open(args_.varfile, "r") as file:
						try:
							param_list = yaml.load(file, Loader=yaml.SafeLoader)
						except yaml.scanner.ScannerError:
							log.write(
								"\nx  Error while reading "
								+ args_.varfile
								+ ". Edit the yaml file and try again (i.e. use ':' instead of '=' to specify variables)"
							)
							sys.exit(
								"\nx  Error while reading "
								+ args_.varfile
								+ ". Edit the yaml file and try again (i.e. use ':' instead of '=' to specify variables)"
							)
			for param in param_list:
				if hasattr(args_, param):
					if getattr(args_, param) != param_list[param]:
						log.write(
							"o  RESET "
							+ param
							+ " from "
							+ str(getattr(args_, param))
							+ " to "
							+ str(param_list[param])
						)
						setattr(args_, param, param_list[param])
					else:
						log.write(
							"o  DEFAULT " + param + " : " + str(getattr(args_, param))
						)
	except UnboundLocalError:  # RAUL: Is it just me or this only happens when the file exists, and ens in .yaml and is empty or does not end in .yaml?
		log.write(
			"\no  The specified yaml file containing parameters was not found! Make sure that the valid params file is in the folder where you are running the code.\n"
		)

	return args_, log


# class for logging
class Logger:
	"""
	Class that wraps a file object to abstract the logging.
	"""

	# Class Logger to writargs.input.split('.')[0] output to a file
	def __init__(self, filein, append, suffix="dat"):
		self.log = open(f"{filein}_{append}.{suffix}", "w")

	def write(self, message):
		"""
		Appends a newline character to the message and writes it into the file.

		Parameters
		----------
		message : str
		   text to be written in the log file.
		"""
		self.log.write(f"{message}\n")

	def fatal(self, message):
		"""
		Writes the message to the file. Closes the file and raises an error exit

		Parameters
		----------
		message : str
		   text to be written in the log file.
		"""
		self.write(message)
		self.finalize()
		raise SystemExit(1)

	def finalize(self):
		"""Closes the file"""
		self.log.close()


# OS utils
def creation_of_dup_csv_csearch(csearch):

	"""
	Generates a pandas.DataFrame object with the appropiate columns for the
	conformational search and the minimization.

	Parameters
	----------
	csearch : str
		Conformational search method. Current valid methods are:
		['rdkit','fullmonte','summ']

	Returns
	-------
	pandas.DataFrame
	"""
	# Boolean aliases from args
	is_rdkit = csearch == "rdkit"
	is_fullmonte = csearch == "fullmonte"
	is_crest = csearch == "crest"
	is_summ = csearch == "summ"

	# column blocks definitions
	base_columns = [
		"Molecule",
		"RDKit-Initial-samples",
		"RDKit-energy-window",
		"RDKit-initial_energy_threshold",
		"RDKit-RMSD-and-energy-duplicates",
		"RDKit-Unique-conformers",
	]
	end_columns_no_min = ["CSEARCH time (seconds)", "Overall charge"]
	fullmonte_columns = [
		"FullMonte-Unique-conformers",
	]
	#'FullMonte-conformers',
	#'FullMonte-energy-window',
	#'FullMonte-initial_energy_threshold',
	#'FullMonte-RMSD-and-energy-duplicates']
	summ_columns = [
		"summ-conformers",
		"summ-energy-window",
		"summ-initial_energy_threshold",
		"summ-RMSD-and-energy-duplicates",
		"summ-Unique-conformers",
	]
	crest_columns = ["Molecule", "crest-conformers"]

	# Check Conformer Search method
	if is_rdkit:
		columns = base_columns
	elif is_fullmonte:
		columns = base_columns + fullmonte_columns
	elif is_summ:
		columns = base_columns + summ_columns
	elif is_crest:
		columns = crest_columns
	else:
		return None
	columns += end_columns_no_min
	return pd.DataFrame(columns=columns)


def creation_of_dup_csv_cmin(cmin):

	"""
	Generates a pandas.DataFrame object with the appropiate columns for the
	conformational search and the minimization.

	Parameters
	----------
	cmin : str
		Minimization method. Current valid methods are:
		['xtb','ani']

	Returns
	-------
	pandas.DataFrame
	"""
	# Boolean aliases from args
	is_xtb = cmin == "xtb"
	is_ani = cmin == "ani"

	# column blocks definitions

	xtb_columns = [
		"xTB-Initial-samples",
		"xTB-energy-window",
		"xTB-initial_energy_threshold",
		"xTB-RMSD-and-energy-duplicates",
		"xTB-Unique-conformers",
	]
	ANI_columns = [
		"ANI-Initial-samples",
		"ANI-energy-window",
		"ANI-initial_energy_threshold",
		"ANI-RMSD-and-energy-duplicates",
		"ANI-Unique-conformers",
	]
	end_columns = ["CMIN time (seconds)", "Overall charge"]

	# Check Minimization Method
	if is_ani:
		columns = ANI_columns
	if is_xtb:  # is_ani and is_xtb will not happen, but this is what was written
		columns = xtb_columns

	columns += end_columns
	return pd.DataFrame(columns=columns)


def move_file(destination, source, file):
	"""
	Moves files from the source folder to the destination folder and creates
	the destination folders when needed.

	Parameters
	----------
	destination : str
		path to the destination folder
	src : str
		path to the source folder
	file : str
		full name of the file (file + extension)
	"""

	destination.mkdir(exist_ok=True, parents=True)
	filepath = source / file
	try:
		filepath.rename(destination / file)
	except FileExistsError:
		filepath.replace(destination / file)


# openbabel utils

# com to xyz to sdf for obabel
def com_2_xyz_2_sdf(input, default_charge, start_point=None):
	"""
	com to xyz to sdf for obabel

	Parameters
	----------
	input : str
		path to the file to convert
	start_point : str, optional
		file(path/name?) to the starting point, by default None

	Returns
	-------
	int?
		charge or None?
	"""
	extension = Path(input).suffix

	if start_point is None:
		if extension in ["com", "gjf", "xyz"]:
			file = Path(input)

	else:
		file = Path(start_point)

	filename = Path.stem

	# Create the 'xyz' file and/or get the total charge
	if (
		extension != "xyz"
	):  #  RAUL: Originally this pointed towards args.input, shouldn't it be to args.file?
		xyz, charge = get_info_input(file)
		xyz_txt = "\n".join(xyz)
		with open(f"{filename}.xyz", "w") as F:
			F.write(f"{len(xyz)}\n{filename}\n{xyz_txt}\n")
	else:
		charge = default_charge

	xyz_2_sdf(f"{filename}.xyz")

	return charge


def xyz_2_sdf(file, parent_dir=None):
	"""
	Creates a .sdf file from a .xyz in the specified directory. If no directory
	is specified then the files are created in the current directory.

	Parameters
	----------
	file : str
		filename and extension of an existing .xyz file
	dir : str or pathlib.Path, optional
		a path to the directory where the .xyz file is located
	"""
	if parent_dir is None:
		parent_dir = Path("")
	else:
		parent_dir = Path(parent_dir)
	mol = next(pybel.readfile("xyz", parent_dir / file))
	ofile = Path(file).stem + ".sdf"
	mol.write("sdf", parent_dir / ofile)


def get_info_input(file):
	"""
	Takes an input file and retrieves the coordinates of the atoms and the
	total charge.

	Parameters
	----------
	file : str or pathlib.Path
		A path pointing to a valid .com or .gjf file

	Returns
	-------
	coordinates : list
		A list of strings (without \\n) that contain the xyz coordinates of the
		.gjf or .com file
	charge : str
		A str with the number corresponding to the total charge of the .com or
		.gjf file
	"""

	with open(file, "r") as input_file:
		input_lines = input_file.readlines()

	_iter = input_lines.__iter__()

	line = ''
	# input for Gaussian calculations
	if file.split('.')[1] in ['com','gjf']:

		# Find the command line
		while "#" not in line:
			line = next(_iter)

		# in case the keywords are distributed in multiple lines
		while len(line.split()) > 0:
			line = next(_iter)

		# pass the title lines
		_ = next(_iter)
		while line:
			line = next(_iter).strip()

		# Read charge and multiplicity
		charge, mult = next(_iter).strip().split()

		# Store the atom types and coordinates until next empty line.
		atoms_and_coords = []
		line = next(_iter).strip()
		while line:
			atoms_and_coords.append(line.strip())
			line = next(_iter).strip()

	# input for ORCA calculations
	if file.split('.')[1] == '.inp':

		# Find the line with charge and multiplicity
		while "* xyz" not in line or "* int" not in line:
			line = next(_iter)

		# Read charge and multiplicity
		charge = line.strip().split()[-2]

		# Store the coordinates until next *
		atoms_and_coords = []
		line = next(_iter).strip()
		while len(line.split()) > 1:
			atoms_and_coords.append(line.strip())
			line = next(_iter).strip()

	return atoms_and_coords, charge


# RDKit Utils
def nci_ts_mol(smi, args, constraints_dist, constraints_angle, constraints_dihedral, name):

	if constraints_dist is not None:
		constraints_dist = [[float(y) for y in x] for x in constraints_dist]
		constraints_dist = np.array(constraints_dist)
	if constraints_angle is not None:
		constraints_angle = [[float(y) for y in x] for x in constraints_angle]
		constraints_angle = np.array(constraints_angle)
	if constraints_dihedral is not None:
		constraints_dihedral = [[float(y) for y in x] for x in constraints_dihedral]
		constraints_dihedral = np.array(constraints_dihedral)

	molsH = []
	mols = []
	for m in smi:
		mols.append(Chem.MolFromSmiles(m))
		molsH.append(Chem.AddHs(Chem.MolFromSmiles(m)))

	for m in molsH:
		AllChem.EmbedMultipleConfs(m,numConfs=1)
	for m in mols:
		AllChem.EmbedMultipleConfs(m,numConfs=1)

	coord = [0.0,0.0,5.0]
	molH = molsH[0]
	for fragment in molsH[1:]:
		offset_3d = Geometry.Point3D(coord[0],coord[1],coord[2])
		molH = Chem.CombineMols(molH, fragment, offset_3d)
		coord[1] += 5
		Chem.SanitizeMol(molH)

	coord = [0.0,0.0,5.0]
	mol = mols[0]
	for fragment in mols[1:]:
		offset_3d = Geometry.Point3D(coord[0],coord[1],coord[2])
		mol = Chem.CombineMols(mol, fragment, offset_3d)
		coord[1] += 5
	mol = Chem.AddHs(mol)
	Chem.SanitizeMol(mol)

	atom_map = []
	for atom in mol.GetAtoms():
		atom_map.append(atom.GetAtomMapNum())

	max_map = max(atom_map)
	for a in mol.GetAtoms():
		if a.GetSymbol() =='H':
			max_map +=1
			a.SetAtomMapNum(int(max_map))

	AllChem.ConstrainedEmbed(mol,molH)
	rdmolfiles.MolToXYZFile(mol, name+'.xyz')

	# for atom in mol.GetAtoms():
		# print(atom.GetAtomMapNum(),atom.GetIdx(),atom.GetSymbol())
	nconstraints_dist = []
	if constraints_dist is not None:

		for i,r in enumerate(constraints_dist):
			# print('r:',r[:2])
			nr = []
			for j,ele in enumerate(r[:2]):
				# print('ele:',  ele)
				for atom in mol.GetAtoms():
					if ele == atom.GetAtomMapNum():
						nr.append(float(atom.GetIdx())+1)
						# print('nr:',  nr)
			nr.append(r[-1])
			# print('nr-final:',  nr)
			nconstraints_dist.append(nr)
		nconstraints_dist  = np.array(nconstraints_dist)
		# print(nconstraints_dist)
		# print('----')
	nconstraints_angle = []
	if constraints_angle is not None:

		for i,r in enumerate(constraints_angle):
			# print('r:',r[:2])
			nr = []
			for j,ele in enumerate(r[:3]):
				# print('ele:',  ele)
				for atom in mol.GetAtoms():
					if ele == atom.GetAtomMapNum():
						nr.append(float(atom.GetIdx())+1)
						# print('nr:',  nr)
			nr.append(r[-1])
			# print('nr-final:',  nr)
			nconstraints_angle.append(nr)
		nconstraints_angle  = np.array(nconstraints_angle)
		# print(nconstraints_angle)
		# print('----')
	nconstraints_dihedral = []
	if constraints_dihedral is not None:
		for i,r in enumerate(constraints_dihedral):
			# print('r:',r[:2])
			nr = []
			for j,ele in enumerate(r[:4]):
				# print('ele:',  ele)
				for atom in mol.GetAtoms():
					if ele == atom.GetAtomMapNum():
						nr.append(float(atom.GetIdx())+1)
						# print('nr:',  nr)
			nr.append(r[-1])
			# print('nr-final:',  nr)
			nconstraints_dihedral.append(nr)
		nconstraints_dihedral  = np.array(nconstraints_dihedral)
		# print(nconstraints_dihedral)
		# print('----')

	# print(constraints_dist[:2])
	# for atom in mol.GetAtoms():
	#     print(atom.GetAtomMapNum(),atom.GetIdx(),atom.GetSymbol())
	#     if constraints_dist is not None:
	#         constraints_dist = np.where(constraints_dist==float(atom.GetAtomMapNum()),float(atom.GetIdx())+1,constraints_dist)
	#     if constraints_angle is not None:
	#         constraints_angle = np.where(constraints_angle==float(atom.GetAtomMapNum()),float(atom.GetIdx())+1,constraints_angle)
	#     if constraints_dihedral is not None:
	#         constraints_dihedral = np.where(constraints_dihedral==float(atom.GetAtomMapNum()),float(atom.GetIdx())+1,constraints_dihedral)
	# rdmolfiles.MolToXYZFile(mol, name+'.xyz')
	# print(nconstraints_dist, nconstraints_angle, nconstraints_dihedral)
	return mol, nconstraints_dist, nconstraints_angle, nconstraints_dihedral


def rules_get_charge(mol, args):
	"""
	Automatically sets the charge for metal complexes

	Parameters
	----------
	mol : [type]
		[description]
	args : [type]
		[description]

	Returns
	-------
	charge : int
		Charge of the system
	"""

	C_group = ["C", "Se", "Ge"]
	N_group = ["N", "P", "As"]
	O_group = ["O", "S", "Se"]
	F_group = ["Cl", "Br", "I"]

	M_ligands, N_carbenes, bridge_atoms, neighbours = [], [], [], []
	charge_rules = np.zeros(len(args.metal_idx), dtype=int)
	neighbours = []
	for atom in mol.GetAtoms():
		# get the neighbours of metal atom and calculate the charge of metal center + ligands
		if atom.GetIdx() in args.metal_idx:
			charge_idx = args.metal_idx.index(atom.GetIdx())
			neighbours = atom.GetNeighbors()
			charge_rules[charge_idx] = args.m_oxi[charge_idx]
			for neighbour in neighbours:
				M_ligands.append(neighbour.GetIdx())
				if neighbour.GetTotalValence() == 4:
					if neighbour.GetSymbol() in C_group:
						carbene_like = False
						bridge_ligand = False
						for inside_neighbour in neighbour.GetNeighbors():
							if inside_neighbour.GetSymbol() in N_group:
								if inside_neighbour.GetTotalValence() == 4:
									for N_neighbour in inside_neighbour.GetNeighbors():
										# this option detects bridge ligands that connect two metals such as M--CN--M
										# we use I since the M is still represented as I at this point
										if N_neighbour.GetSymbol() == "I":
											bridge_ligand = True
											bridge_atoms.append(
												inside_neighbour.GetIdx()
											)
									if not bridge_ligand:
										carbene_like = True
										N_carbenes.append(inside_neighbour.GetIdx())
						if not carbene_like:
							charge_rules[charge_idx] = charge_rules[charge_idx] - 1
				elif neighbour.GetTotalValence() == 3:
					if neighbour.GetSymbol() in N_group:
						charge_rules[charge_idx] = charge_rules[charge_idx] - 1
				elif neighbour.GetTotalValence() == 2:
					if neighbour.GetSymbol() in O_group:
						nitrone_like = False
						for inside_neighbour in neighbour.GetNeighbors():
							if inside_neighbour.GetSymbol() in N_group:
								nitrone_like = True
						if not nitrone_like:
							charge_rules[charge_idx] = charge_rules[charge_idx] - 1

				elif neighbour.GetTotalValence() == 1:
					if neighbour.GetSymbol() in F_group:
						charge_rules[charge_idx] = charge_rules[charge_idx] - 1

	# recognizes charged N and O atoms in metal ligands (added to the first metal of the list as default)
	# this group contains atoms that do not count as separate charge groups (i.e. N from Py ligands)
	if len(neighbours) > 0:
		invalid_charged_atoms = M_ligands + N_carbenes + bridge_atoms
		for atom in mol.GetAtoms():
			if atom.GetIdx() not in invalid_charged_atoms:
				if atom.GetSymbol() in N_group:
					if atom.GetTotalValence() == 4:
						charge_rules[0] = charge_rules[0] + 1
				if atom.GetSymbol() in O_group:
					if atom.GetTotalValence() == 1:
						charge_rules[0] = charge_rules[0] - 1
		return charge_rules
	else:
		# no update in charge as it is an organic molecule
		return [
			args.charge_default,
		]


def substituted_mol(mol, args):
	"""
	Returns a molecule object in which all metal atoms specified in args.metal
	are replaced by Iodine and the charge is set depending on the number of
	neighbors.

	Parameters
	----------
	mol : rdkit.Chem.rdchem.Mol
		A molecule object whose metal is to be substituted
	args : argparse.args
		[description]

	Returns
	-------
	tuple
		mol,args.metal_idx, args.complex_coord, args.metal_sym
	"""
	Neighbors2FormalCharge = dict()
	for i, j in zip(range(2, 9), range(-3, 4)):
		Neighbors2FormalCharge[i] = j

	for atom in mol.GetAtoms():
		symbol = atom.GetSymbol()
		if symbol in args.metal:
			args.metal_sym[args.metal.index(symbol)] = symbol
			args.metal_idx[args.metal.index(symbol)] = atom.GetIdx()
			args.complex_coord[args.metal.index(symbol)] = len(atom.GetNeighbors())
			atom.SetAtomicNum(53)
			n_neighbors = len(atom.GetNeighbors())
			if n_neighbors > 1:
				formal_charge = Neighbors2FormalCharge[n_neighbors]
				atom.SetFormalCharge(formal_charge)

	return mol, args.metal_idx, args.complex_coord, args.metal_sym


# DETECTS DIHEDRALS IN THE MOLECULE
def getDihedralMatches(mol, heavy, log):
	# this is rdkit's "strict" pattern
	pattern = r"*~[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])&!$([CD3](=[N,O,S])-!@[#7,O,S!D1])&!$([#7,O,S!D1]-!@[CD3]=[N,O,S])&!$([CD3](=[N+])-!@[#7!D1])&!$([#7!D1]-!@[CD3]=[N+])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]~*"
	qmol = Chem.MolFromSmarts(pattern)
	matches = mol.GetSubstructMatches(qmol)

	# these are all sets of 4 atoms, uniquify by middle two
	uniqmatches = []
	seen = set()
	for (a, b, c, d) in matches:
		if (b, c) not in seen and (c, b) not in seen:
			if heavy:
				if (
					mol.GetAtomWithIdx(a).GetSymbol() != "H"
					and mol.GetAtomWithIdx(d).GetSymbol() != "H"
				):
					seen.add((b, c))
					uniqmatches.append((a, b, c, d))
			if not heavy:
				if (
					mol.GetAtomWithIdx(c).GetSymbol() == "C"
					and mol.GetAtomWithIdx(d).GetSymbol() == "H"
				):
					pass
				else:
					seen.add((b, c))
					uniqmatches.append((a, b, c, d))
	return uniqmatches


def getDihedralMatches_v2(mol, heavy, log):  # New version using openbabel
	# If this version is selected, bring the import to the top
	# import pybel #
	from openbabel import pybel  # for openbabel>=3.0.0

	AtomInTripleBond = "$(*#*)"
	TerminalAtom = "D1"
	CF3 = "$(C(F)(F)F)"
	CCl3 = "$(C(Cl)(Cl)Cl)"
	CBr3 = "$(C(Br)(Br)Br)"
	tBut = "$(C([CH3])([CH3])[CH3])"
	# A 3-bonded C with a double bond to (N, O or S)
	# singlgy bonded to not ring bonded to a non-terminal N,O or S.
	CD3_1d = "$([CD3](=[N,O,S])-!@[#7,O,S!D1])"
	CD3_1r = "$([#7,O,S!D1]-!@[CD3]=[N,O,S])"  # Backwards version
	# A 3-bonded C with a double bond to (N+)
	# singlgy bonded to not ring bonded to Any non-terminal N
	CD3_2d = "$([CD3](=[N+])-!@[#7!D1])"
	CD3_2r = "$([#7!D1]-!@[CD3]=[N+])"  # Backwards version
	Atom1 = "*"
	Atom2 = f"!{AtomInTripleBond}&!{TerminalAtom}"
	Atom2 += f"&!{CF3}&!{CCl3}&!{CBr3}&!{tBut}"
	Atom2 += f"&!{CD3_1d}&!{CD3_1r}&!{CD3_2d}&!{CD3_2r}"
	Atom3 = f"!{AtomInTripleBond}&!{TerminalAtom}"
	Atom3 += f"&!{CF3}&!{CCl3}&!{CBr3}&!{tBut}"
	Atom4 = "*"
	pattern = f"{Atom1}~[{Atom2}]-!@[{Atom3}]~{Atom4}"
	smarts = pybel.Smarts(pattern)
	matches = smarts.findall(mol)

	# these are all sets of 4 atoms, uniquify by middle two
	H_atoms = set(pybel.Smarts("[#1]").findall(mol))
	C_atoms = set(pybel.Smarts("[#6]").findall(mol))
	uniqmatches = []
	seen = set()
	for (a, b, c, d) in matches:
		if (b, c) not in seen and (c, b) not in seen:
			if heavy:
				if a not in H_atoms and d not in H_atoms:
					seen.add((b, c))
					uniqmatches.append((a, b, c, d))
			if not heavy:
				# So what if a == 'H' and b == 'C'? is that valid ¿?
				if c not in C_atoms or d not in H_atoms:
					seen.add((b, c))
					uniqmatches.append((a, b, c, d))


# checks for salts
def check_for_pieces(smi):
	# taking largest component for salts
	pieces = smi.split(".")
	# if len(pieces) > 1:
	#     # take largest component by length
	#     smi = max(pieces, key=len)
	return pieces


def mol_from_sdf_or_mol_or_mol2(input):
	"""
	mol from sdf

	Parameters
	----------
	input : str
		path to a .sdf .mol or .mol2 file

	Returns
	-------
	tuple of lists?
		suppl, IDs, charges
	"""
	filename = os.path.splitext(input)[0]
	extension = os.path.splitext(input)[1]

	if extension == ".sdf":
		suppl = Chem.SDMolSupplier(input, removeHs=False)
	elif extension == ".mol":
		suppl = Chem.MolFromMolFile(input, removeHs=False)
	elif extension == ".mol2":
		suppl = Chem.MolFromMol2File(input, removeHs=False)

	IDs, charges = [], []

	with open(input, "r") as F:
		lines = F.readlines()

	molecule_count = 0
	for i, line in enumerate(lines):
		if line.find(">  <ID>") > -1:
			ID = lines[i + 1].split()[0]
			IDs.append(ID)
		if line.find("M  CHG") > -1:
			charge_line = line.split("  ")
			charge = 0
			for j in range(4, len(charge_line)):
				if (j % 2) == 0:
					if j == len(charge_line) - 1:
						charge_line[j] = charge_line[j].split("\n")[0]
					charge += int(charge_line[j])
			charges.append(charge)
		if line.find("$$$$") > -1:
			molecule_count += 1
			if molecule_count != len(charges):
				charges.append(0)

	if len(IDs) == 0:
		if extension == ".sdf":
			for i in range(len(suppl)):
				IDs.append(f"{filename}_{i}")
		else:
			IDs.append(filename)
	if len(charges) == 0:
		if extension == ".sdf":
			for _ in suppl:
				charges.append(0)
		else:
			charges.append(0)
	return suppl, IDs, charges


# checks the charge on the smi string
def check_charge_smi(smiles,ts):
	for smi in smiles:
		charge = 0
		for i, smi_letter in enumerate(smi):
			if smi_letter == "+":
				if ts:
					if smi[i + 3] == "]":
						charge += 1
					else:
						charge += int(smi[i + 1])
				else:
					if smi[i + 1] == "]":
						charge += 1
					else:
						charge += int(smi[i + 1])

			elif smi_letter == "-":
				if ts:
					if smi[i + 3] == "]":
						charge -= 1
					else:
						charge -= int(smi[i + 1])
				else:
					if smi[i + 1] == "]":
						charge -= 1
					else:
						charge -= int(smi[i + 1])
	return charge


def set_metal_atomic_number(mol, metal_idx, metal_sym):
	"""
	Changes the atomic number of the metal atoms using their indices.

	Parameters
	----------
	mol : rdkit.Chem.Mol
		rdkit molecule object
	metal_idx : list
		sorted list that contains the indices of the metal atoms in the molecule
	metal_sym : list
		sorted list (same order as metal_idx) that contains the symbols of the
		metals in the molecule.
	"""

	for atom in mol.GetAtoms():
		if atom.GetIdx() in metal_idx:
			re_symbol = metal_sym[metal_idx.index(atom.GetIdx())]
			atomic_number = periodic_table().index(re_symbol)
			atom.SetAtomicNum(atomic_number)


def get_conf_RMS(mol1, mol2, c1, c2, heavy, max_matches_RMSD):
	"""
	Takes in two rdkit.Chem.Mol objects and calculates the RMSD between them.
	(As side efect mol1 is left in the aligned state, if heavy is specified
	the side efect will not happen)

	Parameters
	----------
	mol1 : rdkit.Chem.Mol
		Probe molecule
	mol2 : rdkit.Chem.Mol
		Target molecule. The probe is aligned to the target to compute the RMSD.
	c1 : int
		conformation of mol1 to use for the RMSD
	c2 : int
		conformation of mol2 to use for the RMSD
	heavy : bool
		If True it will ignore the H atoms when computing the RMSD.
	max_matches_RMSD : int
		the max number of matches found in a SubstructMatch()

	Returns
	-------
	float
		Returns the best RMSD found.
	"""
	if heavy:
		mol1 = RemoveHs(mol1)
		mol2 = RemoveHs(mol2)
	return GetBestRMS(mol1, mol2, c1, c2, maxMatches=max_matches_RMSD)


def get_filenames(type, name):
	"""
	Finding the file type to move for analysis
	"""
	files = []
	if type == "output":
		formats = ["*.log", "*.LOG", "*.out", "*.OUT", "*json"]
	elif type == "input":
		formats = ["*.com", "*.gjf"]
	for _, format in enumerate(formats):
		if name is None:
			all_files = enumerate(glob.glob(format))
		else:
			all_files = enumerate(glob.glob(name + format))
		for _, file in all_files:
			if file not in files:
				files.append(file)
	return files


def check_isomerization(
	COORDINATES_com,
	COORDINATES_log,
	atom_types_com,
	atom_types_log,
	vdwfrac,
	covfrac,
	init_csv,
	file,
):
	"""
	Inputs two molecules with the atoms in the same order and checks if any bond
	is too different between them.

	Bonds are considered when the distance between two atoms is smaller than
	either the sum of their adjusted VDW radii or covalent radii (dist = n*R1 + n*R2, where
	n is a user defined parameter).

	Bonds forming part of TSs are removed.

	Parameters
	----------
	COORDINATES_com : list of lists containing atomic coordinates
		Molecule 1 (Theoretically, non-optimized)
	COORDINATES_log : list of lists containing atomic coordinates
		Molecule 2 (Theoretically, optimized)
	atom_types_com : list of atoms
		Molecule 1 (Theoretically, non-optimized)
	atom_types_log : list of atoms
		Molecule 2 (Theoretically, optimized)
	vdwfrac : float
		Fraction of the summed VDW radii (default is 0.5)
	covfrac : float
		Fraction of the summed covalent radii (default is 1.10)
	init_csv : dataframe
		Contains connectivity from the original non-optimized molecules (i.e. saved from CSEARCH)
	file : string
		Filename

	Returns
	-------
	bool
		True there is a clearly distorted bond within the geometries.
	"""

	# load connectivity matrix from the starting points and convert string into matrix
	if not init_csv.empty:
		filename = file.replace("_" + file.split("_")[-1], "")
		init_connectivity_string = init_csv[init_csv["code_name"] == filename][
			"initial_connectiv"
		][0]
		init_connectivity = json.loads(
			init_connectivity_string.replace(".", ",")
			.replace(",]", "],")
			.replace("],]", "]]")
		)
		atom_types_com = init_connectivity[0]

	else:
		init_connectivity = gen_connectivity(
			atom_types_com, COORDINATES_com, vdwfrac, covfrac
		)

	# in case the systems are not the same
	if len(atom_types_log) != len(atom_types_com):
		isomerized = True

	else:
		final_connectivity = gen_connectivity(
			atom_types_log, COORDINATES_log, vdwfrac, covfrac
		)

		# check connectivity differences from initial structure
		diff = final_connectivity - init_connectivity

		# remove bonds involved in TSs from connectivity matrixes
		if not init_csv.empty:
			if "TS_atom_idx" in init_csv.columns:
				ts_atoms = init_csv[init_csv["code_name"] == filename]["TS_atom_idx"][
					0
				].split(",")
				for i, ts_idx in enumerate(ts_atoms):
					for j, ts_idx_2 in enumerate(ts_atoms):
						if j > i:
							diff[int(ts_idx)][int(ts_idx_2)] = 0
							diff[int(ts_idx_2)][int(ts_idx)] = 0

		isomerized = np.any(diff)

	return isomerized


def gen_connectivity(atom_types_conn, COORDINATES_conn, vdwfrac, covfrac):
	"""
	Use VDW radii to infer a connectivity matrix
	"""

	conn_mat = np.zeros((len(atom_types_conn), len(atom_types_conn)))
	for i, elem_i in enumerate(atom_types_conn):
		for j, elem_j in enumerate(atom_types_conn):
			if j > i:
				vdw_ij = bondi()[elem_i] + bondi()[elem_j]
				rcov_ij = rcov()[elem_i] + rcov()[elem_j]
				dist_ij = np.linalg.norm(
					np.array(COORDINATES_conn[i]) - np.array(COORDINATES_conn[j])
				)
				if dist_ij / vdw_ij < vdwfrac or dist_ij / rcov_ij < covfrac:
					conn_mat[i][j] = 1
				else:
					pass

	return conn_mat


def read_file(w_dir, file):
	"""
	Reads through a file and retrieves a list with all the lines.
	"""

	os.chdir(w_dir)
	outfile = open(file, "r")
	outlines = outfile.readlines()

	return outlines, outfile


def output_to_mol(file, format, log):
	"""
	Input an XYZ, LOG or OUT file from QM calculations and converts it into
	a mol object.

	Parameters
	----------
	file : string
		Filename
	format : string
		File format
	log : Logger object
		Writes data regarding the results from QCORR

	Returns
	-------
	mol
		Mol object
	ob_compat
		True if openbabel is installed correctly
	rdkit_compat
		True if RDKit is installed correctly
	"""

	ob_compat = True
	rdkit_compat = True
	try:
		import openbabel as ob
	except (ModuleNotFoundError, AttributeError):
		log.write(
			"\nx  Open Babel is not installed correctly, the geom_rules filter will be disabled"
		)
		ob_compat = False
	try:
		from rdkit.Chem import AllChem as Chem
	except (ModuleNotFoundError, AttributeError):
		log.write(
			"\nx  RDKit is not installed correctly, the geom_rules and check_geom filters will be disabled"
		)
		rdkit_compat = False

	# transforms output file into mol object
	# for input (from com to xyz to mol)
	if format == "xyz":
		cmd_obabel = [
			"obabel",
			"-ixyz",
			os.path.splitext(file)[0] + ".xyz",
			"-omol",
			"-O",
			os.path.splitext(file)[0] + ".mol",
		]
	# for output (from log to mol)
	if format in ["log", "out"]:
		cmd_obabel = [
			"obabel",
			"-ilog",
			os.path.splitext(file)[0] + "." + format,
			"-omol",
			"-O",
			os.path.splitext(file)[0] + ".mol",
		]
	subprocess.run(cmd_obabel)
	mol = Chem.MolFromMolFile(file.split(".")[0] + ".mol")

	return mol, ob_compat, rdkit_compat


def read_energies(file):
	"""
	Parses the energies from sdf files .
	"""
	energies = []
	f = open(file, "r")
	readlines = f.readlines()
	for i, _ in enumerate(readlines):
		if readlines[i].find(">  <Energy>") > -1:
			energies.append(float(readlines[i + 1].split()[0]))
	f.close()
	return energies


def get_name_and_charge(name, charge_data):
	"""
	Function to get name of charge from an SDF file.
	"""

	name_list = name.split("_")

	if "xtb" in name_list or "ani" in name_list:
		if "filter" in name_list:
			name_molecule = name[:-21]
		else:
			name_molecule = name[:-4]
	elif "summ" in name_list:
		if "filter" in name_list:
			name_molecule = name[:-22]
		else:
			name_molecule = name[:-5]
	elif "rdkit" in name_list:
		if "filter" in name_list:
			name_molecule = name[:-23]
		else:
			name_molecule = name[:-6]
	elif "fullmonte" in name_list:
		if "filter" in name_list:
			name_molecule = name[:-27]
		else:
			name_molecule = name[:-10]

	if charge_data is not None:
		for i in range(len(charge_data)):
			if charge_data.loc[i, "Molecule"] == name_molecule:
				charge_com = charge_data.loc[i, "Overall charge"]
			else:
				try:
					suppl = Chem.SDMolSupplier(name + ".sdf", removeHs=False)
				except OSError:
					suppl = False
				if suppl:
					mol = suppl[0]
					charge_com = mol.GetProp("Real charge")
				else:
					charge_com = "Invalid"

		return charge_com

	else:
		return name_molecule


def cclib_atoms_coords(cclib_data):
	'''
	Function to convert atomic numbers and coordinate arrays from cclib into
	a format compatible with QPREP.
	'''
	atom_numbers = cclib_data['atoms']['elements']['number']
	atom_types = []
	per_tab = periodic_table()
	for atom_n in atom_numbers:
		if atom_n < len(per_tab):
			atom_symbol = per_tab[atom_n]
		else:
			atom_symbol = "XX"
		atom_types.append(atom_symbol)

	cartesians_array = cclib_data['atoms']['coords']['3d']
	cartesians = [cartesians_array[i:i + 3] for i in range(0, len(cartesians_array), 3)]

	return atom_types,cartesians



#####################################################.
#        This file stores all the functions         #
#             used for filtering                    #
#####################################################.
from functools import partial

from rdkit.Chem import AllChem as Chem
from rdkit.Chem import rdMolTransforms, Descriptors

from aqme.utils import periodic_table, get_conf_RMS

# Aux functions for passes_Ir_bidentate_x3_rule
def is_carbene_like(neighbours):
    """
    Takes a list of the metal neighbour atoms and returns True if they
    can be considered as if it were a carbene.

    Parameters
    ----------
    neighbors : list
            list of neighbour atoms of the metal.

    Returns
    -------
    bool

    """
    carbene_like = False
    N_group = ["N", "P", "As"]
    for inside_neighbour in neighbours:
        if inside_neighbour.GetSymbol() in N_group:
            if inside_neighbour.GetTotalValence() == 4:
                carbene_like = True
    return carbene_like


def get_fragmented_versions(mol, atom_i, atom_j, metal_idx):
    """
    Creates two copies of the mol object where in the first one the atom_i and
    in the second the atom_j are not bonded to the metal.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
            The molecule to fragment.
    atom_i : int
            Idx of the first atom
    atom_j : int
            Idx of the second atom
    metal_idx : int
            Idx of the metal

    Returns
    -------
    tuple
            mol_i, mol_j
    """
    bond_i = mol.GetBondBetweenAtoms(atom_i, metal_idx)
    mol_i = Chem.FragmentOnBonds(
        mol, [bond_i.GetIdx()], addDummies=True, dummyLabels=[(atom_i, metal_idx)]
    )
    bond_j = mol.GetBondBetweenAtoms(atom_j, metal_idx)
    mol_j = Chem.FragmentOnBonds(
        mol, [bond_j.GetIdx()], addDummies=True, dummyLabels=[(atom_j, metal_idx)]
    )
    return mol_i, mol_j


def is_around_angle(test, angle, offset):
    """
    Checks if a test angle is close to the angle or not.

    Parameters
    ----------
    test : float
            Angle to test in Degrees.
    angle : float
            Angle to compare in Degrees.
    offset : float
            Tolerance around 'angle' in degrees.

    Returns
    -------
    bool
            True if it is in the range [angle-offset,angle+offset].
    """
    return (angle - offset) <= test <= (angle + offset)


def passes_Ir_bidentate_x3_rule_angle_requirements(
    ligand_atoms, mol, metal_idx, offset
):
    """
    Checks if for complexes with Ph_Py ligands if the angles between the N
    atoms are acceptable. Linear when there are 2 Ph_Py and not linear when
    there are 3 Ph_Py Ligands.

    Parameters
    ----------
    ligand_atoms : list
            A list of tuples where the second value of each tuple is the Idx in the
            provided mol of the N atom of that ligand.
    mol : rdkit.Chem.Mol
            The molcule object of the complex
    metal_idx : int
            Idx of the metal atom
    offset : float
            Angle error in degrees to consider or not an angle to be linear.

    Returns
    -------
    bool
            Whether it passes the angle requirements or not
    """
    is_lineal = partial(is_around_angle, angle=180, offset=offset)
    get_angle = lambda i, j: rdMolTransforms.GetAngleDeg(mol, i, metal_idx, j)
    passing = True

    if len(ligand_atoms) == 3:  # For complexes with 3 Ph_Py ligands:
        i, j, k = ligand_atoms
        angles = [get_angle(a[1], b[1]) for a, b in [(i, j), (i, k), (j, k)]]
        passing = not any([is_lineal(angle) for angle in angles])

    if (
        len(ligand_atoms) == 2
    ):  # For complexes with 2 Ph_Py ligands + 1 ligand that is not Ph_Py
        (i, i2), (j, j2) = ligand_atoms
        angle = get_angle(i2, j2)
        passing = is_lineal(angle)

    return passing


def passes_Ir_bidentate_x3_rule(mol, angle_off):
    """
    Checks if a mol containing an Ir metal complex passes the bidentate x3
    rule or not.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
            The molecule to be tested.
    angle_off : float
            Angle in degrees that acts as tolerance around 180º

    Returns
    -------
    bool
            True if it has passed the rule
    """
    bond_threshold = 2.60  # based on observation from DFT optimized geometries
    ligand_links = []
    atom_indexes = []

    passing = True
    # Finds the Ir atom and gets the atom types and indexes of all its neighbours
    # the filter is compatible with molecules that do not contain Ir (always passing)
    metal_idx = None
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() == 77:
            metal_idx = atom.GetIdx()
            for x in atom.GetNeighbors():
                ligand_links.append(x.GetSymbol())
                atom_indexes.append(x.GetIdx())

    # I need to get the only 3D conformer generated in that mol object for rdMolTransforms
    mol_conf = mol.GetConformer(0)
    # This part will identify the pairs of C and N atoms that are part of the same Ph_Py ligand.
    # The shape of the atom pairs is '[[C1_ATOM_NUMBER, N1_ATOM_NUMBER],[C2, N2],...]'.
    # This information is required for the subsequent filtering process based on angles

    # it filters off molecules that the SDF only detects 5 Ir neighbours
    if len(atom_indexes) != 6:
        if metal_idx is not None:
            passing = False
        return passing

    # Filter by distance of the neighbour atoms and the metal
    for atom_idx in atom_indexes:
        bond_length = rdMolTransforms.GetBondLength(mol_conf, metal_idx, atom_idx)
        if bond_length > bond_threshold:
            return False

    ligand_atoms = []
    for atom_i, sym_i in zip(atom_indexes, ligand_links):
        # This is a filter that excludes molecules that fell apart during DFT geometry
        # optimization (i.e. a N atom from one of the ligands separated from Ir). The
        # max distance allowed can be tuned in length_filter
        for atom_j in atom_indexes:
            # Avoid combinations of the same atom with itself
            if atom_i != atom_j and sym_i == "C":
                # We know that the ligands never have 2 carbon atoms bonding the Ir atom except
                # for carbenes. We only use atom_indexes[i] for C atoms, and atom_indexes[j] for the potential
                # N atoms that are part of the same Ph_Py ligand
                neighbours = mol.GetAtoms()[atom_i].GetNeighbors()
                if not is_carbene_like(neighbours):
                    # This part detects the Ir-C bond and breaks it, breaking the Ph_Py ring
                    mol_i, mol_j = get_fragmented_versions(
                        mol, atom_i, atom_j, metal_idx
                    )
                    # identify whether or not the initial 5-membered ring formed between [-Ir-C-C-C-N-] is broken when we break the Ir-C bond. This works
                    # because Ph_Py units bind Ir in the same way always, through 1 C and 1 N that are in the same position, forming a 5-membered ring.
                    # If this ring is broken, atom_indexes[j] will not be part of a 5-membered ring (atom.IsInRingSize(5) == False) which means that
                    # this atom was initially inside the same ligand as the parent C of atom_indexes[i])
                    if not mol_i.GetAtomWithIdx(atom_i).IsInRingSize(5):
                        # doing backwards as well eg. Ir N bond
                        if not mol_j.GetAtomWithIdx(atom_i).IsInRingSize(5):
                            ligand_atoms.append([atom_i, atom_j])
                            break
                    else:
                        if not mol_i.GetAtomWithIdx(atom_j).IsInRingSize(5):
                            if mol.GetAtomWithIdx(atom_j).IsInRingSize(5):
                                ligand_atoms.append([atom_i, atom_j])
                                break
    # Check Angles
    passing = passing and passes_Ir_bidentate_x3_rule_angle_requirements(
        ligand_atoms, mol_conf, metal_idx, angle_off
    )
    return passing


# Main API of the module
def geom_rules_output(mol, args, log, file, print_error_geom_rules):
    """
    returns if a mol object passes all the discarding rules or not.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
            molecule to be tested.
    args : argparse.args
            [description]
    log : Logger
            [description]
    file : str
            Only used to write to the log
    print_error_geom_rules : bool
            Controls extra writing to the log.

    Returns
    -------
    bool
            If True, it means that it is in accordance with the rules
    """
    passing = True
    for rule in args.geom_rules:
        if rule == "Ir_bidentate_x3":
            passing = passes_Ir_bidentate_x3_rule(mol, args.angle_off)
        else:
            var = rule.split(",")
            if len(var) < 2:
                log.write(
                    "x  The geom_rules parameter(s) was not correctly defined, this filter will be turned off"
                )
                return True
            atoms_filter = var[0].split("-")
            angle_rules = int(var[1])
            # the elements of this initial list will be replaced by the corresponding atom id numebrs
            atom_idx = ["ATOM1", "ATOM2", "ATOM3"]

            find_angle = 0
            incompatibility_found = False
            for atom in mol.GetAtoms():
                # count matches
                neigh_count_first = 0
                neigh_count_second = 0
                sym_0, sym_1, sym_2 = atoms_filter
                # Finds the metal atom and gets the atom types and indexes of all its neighbours
                if atom.GetSymbol() == sym_1:
                    # idx of the central atom
                    atom_idx[1] = atom.GetIdx()
                    for x in atom.GetNeighbors():
                        sym = x.GetSymbol()
                        if sym == sym_0 or sym == sym_2:
                            # this ensures that both neighbours are used
                            if sym == sym_0 and sym == sym_2:
                                if neigh_count_first <= neigh_count_second:
                                    neigh_count_first += 1
                                    atom_idx[0] = x.GetIdx()
                                else:
                                    neigh_count_second += 1
                                    atom_idx[2] = x.GetIdx()
                            elif sym == sym_0:
                                neigh_count_first += 1
                                atom_idx[0] = x.GetIdx()
                            elif sym == sym_0:
                                neigh_count_second += 1
                                atom_idx[2] = x.GetIdx()
                            # count matches
                            matches = neigh_count_first + neigh_count_second
                            if matches > 2:
                                if not print_error_geom_rules:
                                    log.write(
                                        f"x  There are multiple options in geom_rules for {file}, this filter will be turned off"
                                    )
                                    incompatibility_found = True
                                    break
                    if neigh_count_first == 1 and neigh_count_second == 1:
                        find_angle += 1
            if find_angle == 0 and not incompatibility_found:
                if not print_error_geom_rules:
                    log.write(
                        f"x  No angles matching the description from geom_rules in {file}, this filter will be turned off"
                    )
            elif find_angle > 1:
                log.write(
                    f"x  {file} contain more than one atom that meets the geom_rules criteria, this filter will be turned off"
                )
            elif find_angle == 1:
                # I need to get the only 3D conformer generated in that mol object for rdMolTransforms
                mol_conf = mol.GetConformer(0)
                # Calculate the angle between the 3 elements
                angle = rdMolTransforms.GetAngleDeg(
                    mol_conf, atom_idx[0], atom_idx[1], atom_idx[2]
                )
                passing = not is_around_angle(angle, angle_rules, args.angle_off)
            if not passing:
                break
    return passing


def filters(mol, log, molwt_cutoff, verbose):
    """
    Applies some basic filters (molwt, salts[currently off], weird atom symbols)
    that only require SMILES data from a compound and returns if the molecule
    passes the filters or not.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
            [description]
    args : argparse.args
            [description]
    log : Logger
            [description]
    molwt_cutoff : float
            Maximum Acceptable Molecular weight.
    verbose : bool
            Controls extra printing in the log. If True writing to the log will be
            enabled otherwise no logging will be produced from this function.


    Returns
    -------
    bool
            True if it passes the filters, False if it should be discarded.
    """

    # Filter 1
    if Descriptors.MolWt(mol) >= molwt_cutoff:
        if verbose:
            log.write(f" Exiting as total molar mass > {molwt_cutoff}")
        return False

    # Filter 2
    # if len(Chem.MolToSmiles(mol).split('.')) > 1: # Disconnected Molecule
    #    return False

    # Filter 3
    symbols = [atom.GetSymbol() for atom in mol.GetAtoms()]

    unknown_atoms = list(set(symbols) - set(periodic_table()))
    if unknown_atoms:
        if verbose:
            log.write(
                f" Exiting as atoms [{','.join(unknown_atoms)}] are not in the periodic table"
            )
        return False

    # Passed
    return True


def ewin_filter(
    sorted_all_cids,
    cenergy,
    args,
    dup_data,
    dup_data_idx,
    log,
    calc_type,
    energy_window,
):
    """
    Given a sorted list of Compound Ids and a sorted list of their energies
    it discards all compound Ids that have an energy higher than the
    args.ewin_csearch with respect to the lowest one.

    Parameters
    ----------
    sorted_all_cids : list
            [description]
    cenergy : list
            [description]
    args : argparse.args
            [description]
    dup_data : pd.Dataframe
            [description]
    dup_data_idx : pd.Dataframe?
            [description]
    log : Logger
            [description]
    calc_type : str
            A string that points towards the column of the dataframe that should
            be filled with the number of duplicates. The current choices are:
            ['rdkit','summ','ani','xtb']
    energy_window : float
            Minimum energy difference with respect to the lowest compound
            discard a compound.

    Returns
    -------
    list
            list of cids accepted
    """
    verbose = args.verbose

    sortedcids = []
    count = 0

    cenergy_min = cenergy[sorted_all_cids[0]]
    # Filter by Energy Window
    for cid in sorted_all_cids:
        if abs(cenergy[cid] - cenergy_min) < energy_window:
            sortedcids.append(cid)
        else:
            count += 1

    log_msg = ""
    if calc_type == "rdkit":
        key = "RDKit"
    elif calc_type == "summ":
        key = "summ"
    elif calc_type == "ani":
        key = "ANI"
    elif calc_type == "xtb":
        key = "xTB"

    # Write it
    dup_data.at[dup_data_idx, f"{key}-energy-window"] = count

    # Log it
    if verbose:
        msg = "o  {} conformers rejected {}based on energy window ewin_csearch (E > {} kcal/mol)"
        log.write(msg.format(count, log_msg, energy_window))

    return sortedcids


def pre_E_filter(
    sortedcids, cenergy, dup_data, dup_data_idx, log, calc_type, threshold, verbose
):
    """
    This filter selects the first compound that it finds with energy an energy
    difference higher or equal to the threshold with respect to the previously
    admitted compounds. (Thought as filter for rdkit)

    Parameters
    ----------
    sortedcids : list or pd.Dataframe?
            List of compound Ids.
    cenergy : list or pd.Dataframe?
            list of compound energies
    dup_data : pd.Dataframe
            [description]
    dup_data_idx : pd.Dataframe?
            [description]
    log : Logger
            [description]
    calc_type : str
            A string that points towards the column of the dataframe that should
            be filled with the number of duplicates. The current choices are:
            ['rdkit','summ','ani','xtb']
    threshold : float
            Minimum energy difference to consider two compounds as different.
            (kcal/mol)

    Returns
    -------
    list
            list of accepted compound Ids
    """
    selectedcids_initial = []
    eng_dup = 0
    # bar = IncrementalBar('o  Filtering based on energy (pre-filter)', max = len(sortedcids))

    # Add the first one
    selectedcids_initial.append(sortedcids[0])
    for conf in sortedcids[1:]:
        is_unique = True
        # check rmsd
        for seenconf in selectedcids_initial:
            E_diff = abs(cenergy[conf] - cenergy[seenconf])  # in kcal/mol
            if E_diff < threshold:
                eng_dup += 1
                is_unique = False
                break
        if is_unique:
            if conf not in selectedcids_initial:
                selectedcids_initial.append(conf)
        # bar.next()
    # bar.finish()

    if verbose:
        log.write(
            f"o  {str(eng_dup)} duplicates removed  pre-energy filter (E < {threshold} kcal/mol)"
        )

    if calc_type == "rdkit":
        column = "RDKit-initial_energy_threshold"
    elif calc_type == "summ":
        column = "summ-initial_energy_threshold"
    # if calc_type == 'fullmonte': column = 'FullMonte-initial_energy_threshold'
    elif calc_type == "ani":
        column = "ANI-initial_energy_threshold"
    elif calc_type == "xtb":
        column = "xTB-initial_energy_threshold"
    else:
        column = ""

    if column:
        dup_data.at[dup_data_idx, column] = eng_dup
    return selectedcids_initial


def RMSD_and_E_filter(
    outmols, selectedcids_initial, cenergy, args, dup_data, dup_data_idx, log, calc_type
):
    """
    This filter selects the first compound that it finds with energy an energy
    difference lower than the threshold with a higher than the threshold rms
    with respect to the nearest (in energy) accepted compound.

    Parameters
    ----------
    outmols : list
            list of rdkit.Chem.Mol objects that can be accesed sorted by CIDs
    selectedcids_initial : list
            list of Compound Ids to filter.
    cenergy : list
            list of energies of the molecules sorted by CIDs
    args : argparse.args
            [description]
    dup_data : pd.Dataframe
            [description]
    dup_data_idx : [type]
            [description]
    log : Logger
            [description]
    calc_type : str
            A string that points towards the column of the dataframe that should
            be filled with the number of duplicates. The current choices are:
            ['rdkit','summ','ani','xtb']

    Returns
    -------
    list
            list of accepted compound Ids
    """
    rms_threshold = args.rms_threshold
    energy_threshold = args.energy_threshold
    verbose = args.verbose
    heavyonly = args.heavyonly
    max_matches_RMSD = args.max_matches_RMSD

    if calc_type not in ["rdkit", "summ", "xtb", "ani"]:
        # RAUL: If this is true "ALL conformers are going to be written"
        #       I Think this should be taken care at the beggining of the
        #       function or before using the function.
        pass

    if verbose:
        log.write(
            f"o  Removing duplicate conformers (RMSD < {rms_threshold} and E difference < {energy_threshold} kcal/mol)"
        )
    # bar = IncrementalBar('o  Filtering based on energy and RMSD', max = len(selectedcids_initial))

    selectedcids = []
    eng_rms_dup = 0
    selectedcids.append(selectedcids_initial[0])

    for i, conf in enumerate(selectedcids_initial[1:]):
        # This keeps track of whether or not your conformer is unique
        excluded_conf = False

        # check energy and rmsd
        for seenconf in selectedcids:
            E_diff = abs(cenergy[conf] - cenergy[seenconf])  # in kcal/mol
            if E_diff < args.energy_threshold:
                if calc_type == "rdkit":
                    rms = get_conf_RMS(
                        outmols[seenconf],
                        outmols[conf],
                        seenconf,
                        conf,
                        args.heavyonly,
                        args.max_matches_RMSD,
                    )
                # elif calc_type == 'summ' or calc_type == 'fullmonte' or calc_type =='xtb' or calc_type =='ani':
                elif calc_type == "summ" or calc_type == "xtb" or calc_type == "ani":
                    rms = get_conf_RMS(
                        outmols[conf],
                        outmols[seenconf],
                        -1,
                        -1,
                        args.heavyonly,
                        args.max_matches_RMSD,
                    )
                if rms < args.rms_threshold:
                    excluded_conf = True
                    eng_rms_dup += 1
                    break
        if not excluded_conf:
            if conf not in selectedcids:
                selectedcids.append(conf)
        # bar.next()
    # bar.finish()

    if verbose:
        log.write(
            f"o  {eng_rms_dup} duplicates removed (RMSD < {rms_threshold} / E < {energy_threshold} kcal/mol)"
        )
        log.write(f"o  {len(selectedcids)} unique conformers remain")

    # Write the found duplicates:
    if calc_type == "rdkit":
        key = "RDKit"
    elif calc_type == "summ":
        key = "summ"
    elif calc_type == "ani":
        key = "ANI"
    elif calc_type == "xtb":
        key = "xTB"
    else:
        key = ""

    if key:
        duplicates_column = f"{key}-RMSD-and-energy-duplicates"
        uniques_column = f"{key}-Unique-conformers"
        dup_data.at[dup_data_idx, duplicates_column] = eng_rms_dup
        dup_data.at[dup_data_idx, uniques_column] = len(selectedcids)

    return selectedcids


# Base classes for the filters (Raul's refactoring part, work in progress)

# class Filter(object):
#     """
#     Base class for the definition of different types of filters. Defines the
#     basic API that any Filter object should have.

#     Parameters
#     ----------
#     function : Function
#             A single parameter function that returns true when the item should pass
#             the filter and False otherwise.

#     Attributes
#     ----------
#     dataset : list or None
#             A list containing a reference to all the elements that where filtered.
#     outcomes : list or None
#             A list of booleans with the same order as the dataset elements.
#     discarded : list or None
#             A list of discarded items from the dataset.
#     accepted : list or None
#             A list of accepted items from the dataset.
#     """

#     def __init__(self, function=None):
#         if function is not None:
#             self.function = function
#         elif getattr(self, "function", None) is None:
#             self.function = lambda x: True
#         self.dataset = None
#         self.outcomes = None
#         self._discarded = None
#         self._accepted = None

#     def add_dummy_parameter(self, function):
#         """
#         Adds a dummy parameter as the first positional parameter to a given
#         function and returns the wrapped function.
#         """

#         def f(dummy, parameter):
#             return function(parameter)

#         return f

#     def apply(self, dataset, key=None, force=False):
#         """
#         Applies the filter to an iterable. Setting the 'dataset' and 'outcomes'
#         attributes of the Filter in the process.

#         Parameters
#         ----------
#         dataset : iterable
#                 Iterable that contains the items to run the filtering
#         key : [type], optional
#                 [description], by default None
#         force : bool, optional
#                 A True value will apply the filter to the dataset forcefully,
#                 overwriting any previous usage of the filter to other dataset.

#         Raises
#         ------
#         ValueError
#                 If the dataset has been already applied to a dataset and the force
#                 keyword is set to 'false'
#         """
#         if self.dataset is not None and not force:
#             msg = "Attempting to apply a previously applied filter with force='false'"
#             raise ValueError(msg)
#         elif self.dataset is not None:
#             self.clean()

#         if key is None:
#             key = lambda x: x
#         self.dataset = list(dataset)
#         outcomes = self.calc_outcomes(dataset, key)
#         self.outcomes = tuple(outcomes)

#     def calc_outcomes(self, dataset, key):
#         """
#         Runs the filter on a dataset and returns the outcomes without storing
#         the values.

#         Parameters
#         ----------
#         dataset : iterable
#                 Iterable that contains the items to run the filtering
#         key : function
#                 [description]
#         """
#         return tuple(self.function(key(data)) for data in dataset)

#     def extend(self, dataset, key=None):
#         """
#         Extends the current dataset with the new dataset and includes the
#         outcomes of applying the filter to the new dataset.

#         Parameters
#         ----------
#         dataset : iterable
#                 Iterable that contains the items to run the filtering
#         key : [type], optional
#                 [description], by default None
#         """
#         if key is None:
#             key = lambda x: x
#         new = list(dataset)
#         outcomes = self.calc_outcomes(new, key)

#         self.dataset = self.dataset + new
#         self.outcomes = tuple(i for i in chain(self.outcomes, outcomes))
#         # And resets the cache of discarded and accepted
#         self._accepted = None
#         self._discarded = None

#     def clean(self):
#         """
#         Resets the state of the Filter removing all the information stored
#         about the last application of the of the filter to a set of data.
#         """
#         self.dataset = None
#         self.outcomes = None
#         self._discarded = None
#         self._accepted = None

#     @property
#     def discarded(self):
#         if self._discarded is None and self.dataset is not None:
#             self._discarded = [
#                 d for out, d in zip(self.outcomes, self.dataset) if not out
#             ]
#         return self._discarded

#     @property
#     def accepted(self):
#         if self._accepted is None and self.dataset is not None:
#             self._accepted = [d for out, d in zip(self.outcomes, self.dataset) if out]
#         return self._accepted


# class CompoundFilter(Filter):
#     """
#     Class used to apply several filters to a same dataset in a certain order and
#     store the information about which ones were discarded in which filter.

#     Parameters
#     ----------
#     *filters : Filter
#             An undefined number of Filter objects

#     Attributes
#     ----------
#     filters : list
#             List of Filter objects in application order.
#     dataset : list or None
#             A list containing a reference to all the elements that where filtered.
#     outcomes : list or None
#             A list of booleans with the same order as the dataset elements.
#     discarded : list or None
#             A list of discarded items from the dataset.
#     accepted : list or None
#             A list of accepted items from the dataset.
#     """

#     def __init__(self, *filters):
#         self.filters = filters
#         super().__init__()

#     def insert(self, index, item):
#         """
#         Inserts a filter before the index position.

#         Parameters
#         ----------
#         item : Filter
#                 The filter object to include.
#         """
#         self.filters.insert(index, item)

#     def pop(self, index=-1):
#         """
#         Remove and return a filter at index (default last).
#         """
#         return self.filters.pop(index)

#     def apply(self, dataset, key=None, keys=None, force=False):
#         """
#         Applies the filter to an iterable. Setting the 'dataset' and 'outcomes'
#         attributes of each Filter in the process.

#         Parameters
#         ----------
#         dataset : iterable
#                 Iterable that contains the items to run the filtering
#         key : function, optional
#                 If no function is provided it is assumed that the filter can process
#                 each item in the dataset without any change, by default None.
#         keys : iterable, optional
#                 Iterable of functions with the same length as the number of filters
#                 to provide different, by default None. Overrides the key argument.
#         force : bool, optional
#                 A True value will apply the filter to the dataset forcefully,
#                 overwriting any previous usage of the filter to other dataset.

#         Raises
#         ------
#         ValueError
#                 If the dataset has been already applied to a dataset and the force
#                 keyword is set to 'false'
#         """
#         if self.dataset is not None and not force:
#             msg = "Attempting to apply a previously applied filter with force='false'"
#             raise ValueError(msg)
#         elif any([f.dataset is not None for f in self.filters]) and not force:
#             msg = (
#                 "At least one of the filters has already been applied and force='false'"
#             )
#             raise ValueError(msg)
#         elif force:
#             self.clean()

#         # Assign the keys iterable
#         if keys is None and key is None:
#             key = lambda x: x
#             keys = [key for _ in self.filters]
#         elif key is not None:
#             keys = [key for _ in self.filters]

#         self.dataset = dataset
#         outcomes = self.calc_outcomes(dataset, keys)

#         # Set the attributes for all the Filters
#         self.filters[0].dataset = dataset
#         self.filters[0].outcomes = outcomes[0]
#         out_old = outcomes[0]
#         for f, out in zip(self.filters[1:], outcomes[1:]):
#             dataset_n = [d for i, d in zip(out_old, self.dataset) if i]
#             f.dataset = dataset_n
#             f.outcomes = out
#             out_old = out

#         # Set the attributes for the current object
#         outcomes = self.homogenize_outcomes(outcomes)

#         self.outcomes = outcomes

#     def calc_outcomes(self, dataset, keys):
#         """
#         Runs the filter on a dataset and returns the outcomes without storing
#         the values.

#         Parameters
#         ----------
#         dataset : iterable
#                 Iterable that contains the items to run the filtering
#         keys : iterable
#                 Iterable of functions that ensure proper input per each filter.
#         """
#         outcomes = []
#         dataset_old = dataset
#         # outcomes_old = (True,)*len(dataset)
#         for f, key in zip(self.filters, keys):
#             # Use the filter and get the dataset for the next filter
#             out = f.calc_outcomes(dataset_old, key=key)
#             dataset_old = [d for i, d in zip(out, dataset_old) if i]
#             outcomes.append(tuple(out))
#         return outcomes

#     def homogenize_outcomes(self, outcomes):
#         """
#         Returns a list of tuples where all tuples are the same size.

#         Parameters
#         ----------
#         outcomes : list
#                 List of tuples with the outcomes of each of the filters.
#         n : int
#                 size of the dataset

#         Returns
#         -------
#         list
#                 list of the homogenized tuples.
#         """
#         homogenized = []
#         outcomes_old = (True,) * len(outcomes[0])
#         for _, out in zip(self.filters, outcomes):
#             _iter = out.__iter__()
#             outcomes_new = [False if not i else next(_iter) for i in outcomes_old]
#             homogenized.append(tuple(outcomes_new))
#             outcomes_old = outcomes_new
#         return homogenized

#     def extend(self, dataset, key=None, keys=None):
#         """
#         Extends the current dataset with the new dataset and includes the
#         outcomes of applying the filter to the new dataset.

#         Parameters
#         ----------
#         dataset : iterable
#                 Iterable that contains the items to run the filtering
#         key : function, optional
#                 If no function is provided it is assumed that the filter can process
#                 each item in the dataset without any change, by default None.
#         keys : iterable, optional
#                 Iterable of functions with the same length as the number of filters
#                 to provide different, by default None. Overrides the key argument.
#         """
#         # Assign the keys iterable
#         if keys is None and key is None:
#             key = lambda x: x
#             keys = [key for _ in self.filters]
#         elif key is not None:
#             keys = [key for _ in self.filters]

#         new = list(dataset)

#         _outcomes = self.outcomes
#         outcomes = self.calc_outcomes(new, keys)

#         # Set the attributes for the current object
#         self.dataset = self.dataset + new

#         # reset the cache of the properties
#         self._accepted = None
#         self._discarded = None

#         # Set the attributes for all the Filters
#         self.filters[0].dataset = self.dataset
#         self.filters[0].outcomes = self.filters[0].outcomes + outcomes[0]
#         out_old = outcomes[0]
#         for f, out in zip(self.filters[1:], outcomes[1:]):
#             dataset_n = [d for i, d in zip(out_old, self.dataset) if i]
#             f.dataset = f.dataset + dataset_n
#             f.outcomes = f.outcomes + out
#             # reset the cache of the properties
#             f._accepted = None
#             f._discarded = None
#         outcomes = self.homogenize_outcomes(outcomes)
#         self.outcomes = [
#             out_old + out_new for out_old, out_new in zip(_outcomes, outcomes)
#         ]

#     def clean(self):
#         """
#         Resets the state of the CompoundFilter removing all the information
#         stored about the last application of the of the filter to a set of data.
#         It cleans all the component filters.
#         """
#         super().clean()
#         for f in self.filters:
#             f.clean()

#     def accepted_from(self, index):
#         """
#         returns the list of accepted items at the specified filter.
#         """
#         return [d for out, d in zip(self.outcomes[index], self.dataset) if out]

#     def discarded_from(self, index):
#         """
#         returns the total list of discarded items after the specified filter.
#         """
#         return [d for out, d in zip(self.outcomes[index], self.dataset) if not out]

#     @Filter.discarded.getter
#     def discarded(self):
#         if self._discarded is None and self.dataset is not None:
#             self._discarded = self.discarded_from(-1)
#         return self._discarded

#     @Filter.accepted.getter
#     def accepted(self):
#         if self._accepted is None and self.dataset is not None:
#             self._accepted = self.accepted_from(-1)
#         return self._accepted


# class RMSDFilter(Filter):
#     """
#     This filter inputs tuples of (molecule,cid). Each time a conformer that
#     passes the filter is found it is added to the pool of conformers.

#     Note: The RMSD calculation done by rdkit has the side effect of leaving the
#     target conformer aligned to the probe conformer.

#     Parameters
#     ----------
#     threshold : float
#             Minimum RMSD to accept a conformer.
#     maxmatches : int
#             maximum number of atoms should match in the alignment previous to the
#             RMSD calculation.
#     heavyonly : [type], optional
#             [description], by default True.
#     reverse : bool, optional
#             Reverses the threshold. If True, only conformers with an RMSD < threshold
#             will be accepted. by default False.
#     is_rdkit : bool
#             If the conformers to compare have been generated with rdkit and the
#             cid to use to access the conformer is the one provided instead of -1.
#             by default False.

#     """

#     def __init__(
#         self, threshold, maxmatches, heavyonly=True, reverse=False, is_rdkit=False
#     ):
#         self.threshold = threshold
#         self.maxmatches = maxmatches
#         self.heavyonly = heavyonly
#         self.pool = []
#         self.is_rdkit = is_rdkit
#         super().__init__(function=self.function)

#     def set_initial_pool(self, pool, key=None):
#         """
#         Sets the initial pool of seen conformers.

#         Parameters
#         ----------
#         pool : list
#                 A list of conformers ( or things convertible to conformers through
#                 the key function)
#         key : function, optional
#                 A function to convert each item in the pool to a conformer that
#                 can be subjected to the filter, by default None.
#                 i.e
#                 >>> myconformers[cid<-int] = conformer
#                 >>> pool = [cid1,cid2,...,cidn]
#                 >>> key = lambda x: myconformers[x]
#         """
#         if key is not None:
#             pool = list(map(key, pool))
#         self.pool = pool

#     def clean(self, also_pool=True):
#         """
#         Resets the state of the Filter removing all the information stored
#         about the last application of the of the filter to a set of data.

#         Parameters
#         ----------

#         also_pool : bool
#                 If true it will also clear the initial pool of conformers.
#         """
#         if also_pool:
#             self.pool = []
#         super().__init__()

#     def function(self, item):
#         """
#         main function of the filter. inputs a tuple (molecule,cid) aligns it to
#         all the previously stored molecules and if it passes the filter, stores
#         it and returns True, otherwise returns False.
#         """
#         probe, cid_p = item
#         if item in self.pool:
#             return True
#         for target, cid_t in self.pool:
#             c1 = c2 = -1
#             if self.is_rdkit:
#                 probe, target = target, probe
#                 c1, c2 = cid_t, cid_p
#             rms = get_conf_RMS(probe, target, c1, c2, self.heavyonly, self.maxmatches)
#             if self.reverse:
#                 reject = rms < self.threshold
#             else:
#                 reject = rms > self.threshold
#             if reject:
#                 return False
#         else:
#             self.pool.append(item)
#             return True


# class EnergyFilter(Filter):
#     """
#     This filter inputs energy values. Each time a conformer that
#     passes the filter is found it is added to the pool of conformers.

#     Note: The RMSD calculation done by rdkit has the side effect of leaving the
#     target conformer aligned to the probe conformer.

#     Parameters
#     ----------
#     threshold : float
#             Energy threshold in kcal/mol.
#     mode : str, ['difference','window']
#             'difference' accepts all the conformers whose energy difference
#             with respect to all the previous conformers found is larger than
#             the threshold.
#             'window' accepts all the conformers whose energy difference with respect
#             to the lowest in energy is smaller than the threshold.

#     Attributes
#     ----------
#     pool : list
#             In window mode, it is the list of minima used to calculate the energy
#             window. If the lowest conformer was either set as initial pool or
#             provided as the first item of the dataset the pool will remain of len=1.
#             In difference mode, it corresponds to all the conformers accepted.
#     """

#     def __init__(self, threshold, mode):
#         self.mode = mode  # difference or window
#         self.threshold = threshold
#         self.pool = []
#         super().__init__()

#     def set_initial_pool(self, pool, key=None):
#         """
#         Sets the initial pool of seen conformers.

#         Parameters
#         ----------
#         pool : list
#                 A list of conformers ( or things convertible to conformers through
#                 the key function)
#         key : function, optional
#                 A function to convert each item in the pool to a conformer that
#                 can be subjected to the filter, by default None.
#                 i.e
#                 >>> myconformers[cid<-int] = conformer
#                 >>> pool = [cid1,cid2,...,cidn]
#                 >>> key = lambda x: myconformers[x]
#         """
#         if key is not None:
#             pool = list(map(key, pool))
#         self.pool = pool

#     def clean(self, also_pool=True):
#         """
#         Resets the state of the Filter removing all the information stored
#         about the last application of the of the filter to a set of data.

#         Parameters
#         ----------

#         also_pool : bool
#                 If true it will also clear the initial pool of conformers.
#         """
#         if also_pool:
#             self.pool = []
#         super().__init__()

#     def function(self, item):
#         """
#         main function of the filter. Inputs a tuple (cid,energy) and accepts or
#         rejects the item depending on the energy threshold and filter mode.
#         """
#         assert self.mode in ["window", "difference"]
#         if self.mode == "window":
#             out = self._window(item)
#         else:
#             out = self._difference(item)
#         return out

#     def _difference(self, item):
#         cid_p, energy_p = item
#         if item in self.pool:
#             return True
#         for cid_t, energy_t in self.pool:
#             if abs(energy_p - energy_t) < self.threshold:
#                 return False
#         else:
#             self.pool.append(item)
#             return True

#     def _window(self, item):
#         if not self.pool:  # Ensure the pool has len = 1
#             self.pool.append(item)
#             return True
#         cid_p, energy_p = item
#         cid_t, energy_t = self.pool[-1]
#         reject = abs(energy_p - energy_t) < self.threshold
#         is_lowest = energy_p < energy_t
#         if reject:
#             return False
#         if is_lowest:
#             self.pool.append(item)
#         return True



#####################################################.
#        This file stores all the functions         #
#     used for genrating details for fullmonte      #
#####################################################.

import sys
import numpy as np
import math
import random
from rdkit.Chem import AllChem as Chem
from rdkit.Chem import rdMolTransforms, rdMolAlign

from aqme.utils import set_metal_atomic_number, get_conf_RMS

def realign_mol(mol,conf,coord_Map, alg_Map, mol_template, maxsteps):  # RAUL: This function requires a clear separation between minimization and alignment.
    """
    Minimizes and aligns the molecule provided freezing the atoms that match 
    the mol_template.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
        Molecule to be minimized and aligned. 
    conf : int?
        Number that indicates which conformation of the molecule will be 
        minimized and aligned. 
    coord_Map : [type]
        [description]
    alg_Map : [type]
        [description]
    mol_template : [type]
        [description]
    maxsteps : int
        Maximum number of iterations in FF minimization.

    Returns
    -------
    mol,energy
        The updated mol object and the final forcefield energy.
    """
    num_atom_match = mol.GetSubstructMatch(mol_template)
    forcefield = Chem.UFFGetMoleculeForceField(mol,confId=conf)
    for i, idxI in enumerate(num_atom_match):
        for idxJ in num_atom_match[i+1:]:
            d = coord_Map[idxI].Distance(coord_Map[idxJ])
            forcefield.AddDistanceConstraint(idxI, idxJ, d, d, 10000)
    forcefield.Initialize()
    forcefield.Minimize(maxIts=maxsteps)
    # rotate the embedded conformation onto the core_mol:
    rdMolAlign.AlignMol(mol, 
                        mol_template, 
                        prbCid=conf,
                        refCid=-1,
                        atomMap=alg_Map,
                        reflect=True,
                        maxIters=100)
    energy = float(forcefield.CalcEnergy())
    return mol,energy

def minimize_rdkit_energy(mol,conf,log,FF,maxsteps):
    """
    Minimizes a conformer of a molecule and returns the final energy.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
        [description]
    conf : int?
        Number that indicates which conformation of the molecule will be 
        minimized and aligned.
    args : [type]
        [description]
    log : [type]
        [description]
    FF : str
        A valid forcefield name. Currently only 'UFF' and 'MMFF' supported.
    maxsteps : int
        Maximum number of iterations in FF minimization.

    Returns
    -------
    float
        Final energy of the minimization
    """

    if FF == "MMFF":
        properties = Chem.MMFFGetMoleculeProperties(mol)
        forcefield = Chem.MMFFGetMoleculeForceField(mol,properties,confId=conf)

    if FF == "UFF" or forcefield is None:
        # if forcefield is None means that MMFF will not work. Attempt UFF.
        forcefield = Chem.UFFGetMoleculeForceField(mol,confId=conf)

    if FF not in ["MMFF","UFF"] or forcefield is None:
        log.write(f' Force field {FF} not supported!')
        sys.exit()

    # RAUL: If UFF fails to be set up -> 'forcefield is None'
    #       Should we log it and exit?
    forcefield.Initialize()
    forcefield.Minimize(maxIts=maxsteps)
    energy = float(forcefield.CalcEnergy())

    return energy

def rotate_dihedrals(conformer,dihedrals,seed,stepsize):
    """
    Applies a random rotation to all the dihedrals  

    Parameters
    ----------
    conformer : rdkit.Chem.rdchem.Conformer
        The conformer whose angles are going to be rotated. 
        (conformer = mol.GetConformer(cid)) 
    dihedrals : list
        A list of tuples of all the dihedrals that are going to be rotated. 
    seed : int
        seed for the random module
    stepsize : float
        Angle in Degrees to do the steps between 0.0 and 360.0
    """
    rad_range = np.arange(0.0, 360.0,stepsize)
    for dihedral in dihedrals:
        random.seed(seed)                                                       # RAUL: Any good reason to keep reseting the seed ? 
        rad_ang = random.choice(rad_range)
        rad = math.pi*rad_ang/180.0
        rdMolTransforms.SetDihedralRad(conformer,*dihedral,value=rad)

def generating_conformations_fullmonte(name,args,rotmatches,log,selectedcids_rdkit,outmols,sdwriter,dup_data,dup_data_idx,coord_Map,alg_Map, mol_template):

    ##working with fullmonte
    n_unique_conformers = len(selectedcids_rdkit)
    log.write(f"\n\no  Generation of confomers using FULLMONTE using "  \
              f"{n_unique_conformers} unique conformer(s) as starting point(s)")

    #Writing the conformers as mol objects to sdf
    sdtemp = Chem.SDWriter(name+'_'+'rdkit'+args.output)
    for conf in selectedcids_rdkit:
        sdtemp.write(outmols[conf],conf)
    sdtemp.close()

    fmmols = Chem.SDMolSupplier(name+'_'+'rdkit'+args.output, removeHs=False)
    if fmmols is None:
        log.write("Could not open "+ name+args.output)
        sys.exit(-1)


    # array for each each unique from rdkit
    unique_mol,c_energy,unique_mol_sample  = [],[],[]

    #STEP 1: Use start conformation for and append to unique list
    nsteps = 1
    for mol_fm in fmmols:
        unique_mol.append(mol_fm)
        c_energy.append(float(mol_fm.GetProp("Energy")))

    #defining unique mol sample for choosing
    globmin = min(c_energy)
    for ene in reversed(c_energy):
        if abs(globmin-ene) < args.ewin_sample_fullmonte:
            unique_mol_sample.append(unique_mol[c_energy.index(ene)])

    # bar = IncrementalBar('o  Generating conformations for Full Monte', max = args.nsteps_fullmonte)
    while nsteps < args.nsteps_fullmonte+1:
        
        seed = nsteps

        #STEP 2: Choose mol object form unique_mol:
        random.seed(seed)
        mol_rot = random.choices(unique_mol_sample,k=1)[0]

        #updating the location of mol object i.e., the hexadecimal locaiton to a new one so the older one isnt affected
        mol = Chem.RWMol(mol_rot)
        rot_mol = mol.GetMol()

        #STEP 3: Choose random subset of dihedral from rotmatches
        random.seed(seed)                                                       # RAUL: Any good reason to keep reseting the seed ? 
        k = min(len(rotmatches),args.nrot_fullmonte)
        mutable_dihedrals = random.choices(rotmatches, k=k)
        
        #STEP 4: for the given conformation, then apply a random rotation to each torsion in the subset
        conformer = rot_mol.GetConformer(conf)
        rot_mol = rotate_dihedrals(conformer,mutable_dihedrals,seed,args.ang_fullmonte)

        #STEP 5: Optimize geometry rot_mol
        if (coord_Map, alg_Map, mol_template) == (None, None, None):
            energy = minimize_rdkit_energy(rot_mol,-1,log,args.ff,args.opt_steps_RDKit)
        else:
            mol,energy = realign_mol(rot_mol,-1,coord_Map, alg_Map, mol_template,args.opt_steps_RDKit)

        #STEP 6 : Check for DUPLICATES - energy and rms filter (reuse)
                 #  if the conformer is unique then save it the list
        exclude_conf= False
        # compare against allprevious conformers located
        for j,seenmol in enumerate(unique_mol):
            if abs(energy - c_energy[j]) < args.initial_energy_threshold:
                exclude_conf = True
                break
            if  abs(energy - c_energy[j]) < args.energy_threshold:
                rms = get_conf_RMS(rot_mol,seenmol,-1,-1, args.heavyonly, args.max_matches_RMSD)
                if rms < args.rms_threshold:
                    exclude_conf = True
                    break
        if not exclude_conf:
            unique_mol.append(rot_mol)
            c_energy.append(energy)
            unique_mol[c_energy.index(energy)].SetProp("Energy", str(energy))

        unique_mol_sample = []
        #STEP 7: ANALYSE THE UNIQUE list for lowest energy, reorder the uniques if greater the given thershold remove
        globmin = min(c_energy)
        for ene in reversed(c_energy):
            indx = c_energy.index(ene)
            if abs(globmin-ene) > args.ewin_fullmonte:
                unique_mol.pop(indx)
                c_energy.pop(indx)
            if abs(globmin-ene) < args.ewin_sample_fullmonte:

                unique_mol_sample.append(unique_mol[indx])

        nsteps += 1
    #     bar.next()
    #
    # bar.finish()

    dup_data.at[dup_data_idx, 'FullMonte-Unique-conformers'] = len(unique_mol)

    if args.verbose:
        log.write("o  "+ str(len(unique_mol))+" unique conformers remain")

    cids = list(range(len(unique_mol)))
    sorted_all_cids = sorted(cids,key = lambda cid: c_energy[cid])

    #STEP 9: WRITE FINAL uniques to sdf for xtb or ani
    for i, cid in enumerate(sorted_all_cids):
        unique_mol[cid].SetProp('_Name',name+' '+str(i))
        if coord_Map is None and alg_Map is None and mol_template is None:
            if args.metal_complex:
                set_metal_atomic_number(unique_mol[cid],args.metal_idx,args.metal_sym)
            sdwriter.write(unique_mol[cid])
        else:
            mol_realigned,_ = realign_mol(unique_mol[cid],-1,coord_Map, alg_Map, mol_template,args.opt_steps_RDKit)
            if args.metal_complex:
                set_metal_atomic_number(mol_realigned,args.metal_idx,args.metal_sym)
            sdwriter.write(mol_realigned)

    status = 1

    return status



#####################################################.
#        This file stores all the functions         #
#    used in template based conformer generation    #
#####################################################.
import sys
from pathlib import Path
from pkg_resources import resource_filename

from rdkit.Chem import AllChem as Chem
from rdkit.Chem import rdDistGeom, rdMolAlign

from aqme.utils import get_conf_RMS

TEMPLATES_PATH = Path(resource_filename("aqme", "templates"))

# Auxiliar functions of this module
def load_template(complex_type, log):
	"""
	Checks if the templates are reachable and if so returns the molecule object.

	Returns
	-------
	rdkit.Chem.Mol
		The molecule file of the corresponding template.
	"""
	type2template = dict()
	type2template["squareplanar"] = "template-4-and-5.sdf"
	type2template["squarepyramidal"] = "template-4-and-5.sdf"
	type2template["linear"] = "template-2.sdf"
	type2template["trigonalplanar"] = "template-3.sdf"

	folder = TEMPLATES_PATH

	if not folder.exists():
		log.write(
			"x The templates folder was not found, probably due to a problem while installing AQME"
		)
		sys.exit()

	file_template = folder / Path(type2template[complex_type])
	templates = Chem.SDMolSupplier(str(file_template))
	template = templates[
		-1
	]  # RAUL: I'm assuming that there is only one molecule per template and in case there are several, it's the last one

	return template


def calc_neighbours(molecule, metals_idx):
	"""
	Changes the atomic number (and charge) of the first metal found
	and returns its neighbours, the number of neighbours and the idx of the
	metal.

	Parameters
	----------
	molecule : rdkit.Chem.Mol
		[description]
	metals_idx : list
		List containing the Idx of the metals. The first match is the only one
		considered.

	Returns
	-------
	list
		list of neighbour atoms
	"""
	bonds2AtNum = dict()
	bonds2AtNum[5] = 14
	bonds2AtNum[4] = 14
	bonds2AtNum[3] = 53
	bonds2AtNum[2] = 53
	# RAUL: Are we interested in the first Metal found in the molecule or do we
	#       know beforehand the idx of the Metal atom?. If we know beforehand:
	# idx = metals_idx[0]
	# atom = molecule.GetAtomWithIdx(idx)
	# n_bonds = len(atom.GetBonds())
	# atom.SetAtomicNum(bonds2AtNum[n_bonds])
	# if n_bonds == 5:
	#    atom.SetFormalCharge(1)
	# neighbours = atom.GetNeighbors()
	# return idx, neighbours
	for atom in molecule.GetAtoms():
		idx = atom.GetIdx()
		if idx in metals_idx:
			n_bonds = len(atom.GetBonds())
			AtNum = bonds2AtNum[n_bonds]
			atom.SetAtomicNum(AtNum)
			if n_bonds == 5:
				atom.SetFormalCharge(1)
			neighbours = atom.GetNeighbors()
			return neighbours
	return []


def get_mappings(molecule, template, conformer_id=-1):
	match = molecule.GetSubstructMatch(template)
	conformer = template.GetConformer(conformer_id)
	coordMap = {}
	algMap = []
	for i, atomidx in enumerate(match):
		algMap.append((atomidx, i))
		coordMap[atomidx] = conformer.GetAtomPosition(i)
	return coordMap, algMap


def get_distance_constrains(coordMap):
	atom_idxs = list(coordMap.keys())
	constrains = []
	for k, i in enumerate(atom_idxs):
		for j in atom_idxs[k + 1 :]:
			d = coordMap[i].Distance(coordMap[j])
			constrains.append((i, j, d))
	return constrains


# Auxiliar functions for minimization and filtering
def template_embed_optimize(target, template, maxsteps, log):
	"""
	Embeds a new conformation into a molecule, optimizes it using UFF and
	realigns it.

	Parameters
	----------
	target : rdkit.Chem.Mol
		Molecule where you want to embed the new conformation
	mol_template : rdkit.Chem.Mol?
		Template molecule to identify the core of the molecule that will have
		its distances frozen in the optimization.
	maxsteps : int
		Number of maximum optimization steps in RDKit.
	log : Logger
		[description]

	Returns
	-------
	molecule, coordMap, algMap, conf_id
		molecule embedded, mapping to the atom instances,
		list of tuples with position and id and int with the conformer id.
	"""

	seed = -1
	force_constant = 10000

	coord_map, alg_map = get_mappings(target, template, conformer_id=-1)

	# add H's to molecule
	molecule = Chem.AddHs(target)

	conf_id = rdDistGeom.EmbedMolecule(molecule, coordMap=coord_map, randomSeed=seed)

	if conf_id < 0:
		log.write("Could not embed molecule.")
		return molecule, None, None, conf_id

	forcefield = Chem.UFFGetMoleculeForceField(molecule, confId=conf_id)

	constraints = get_distance_constrains(coord_map)
	for i, j, d in constraints:
		forcefield.AddDistanceConstraint(i, j, d, d, force_constant)
	forcefield.Initialize()
	forcefield.Minimize(maxIts=maxsteps)
	# rotate the embedded conformation onto the core_mol:
	rdMolAlign.AlignMol(molecule, template, atomMap=alg_map, reflect=True, maxIters=100)

	return molecule, coord_map, alg_map, conf_id


def filter_template_mol(molecule_new, mol_objects, heavyonly, max_matches):
	"""
	Returns if a mol object should be kept or not.

	Parameters
	----------
	molecule_new : [type]
		[description]
	mol_objects : [type]
		[description]
	heavyonly : bool
		If True only non-H atoms are considered for the RMS calculation
	max_matches : int
		Maximum number of matches in the RMSD?

	Returns
	-------
	bool
		Returns True when the molecule should be kept and False when it should
		be discarded.
	"""

	if not mol_objects:
		return True

	# check if molecule also exixts in the mol_objects
	for mol in mol_objects:
		rms = get_conf_RMS(mol, molecule_new, -1, -1, heavyonly, max_matches)
		if rms < 0.5:
			return False
	return True


# Decorators for documentation
def doc_parameters(f):
	"""
	Decorator that adds the "Parameters" section at the end of the
	docstrings of the decorated function. Care to use this decorator 'below' the
	doc_returns decorator.

	Parameters
	----------
	f : function
		function to decorate.

	Returns
	-------
	function
		returns the same function with the docstring modified.
	"""
	description = f.__doc__
	parameters = [
		("molecule", "rdkit.Chem.Mol", "Molecule to be embedded "),
		("template", "rdkit.Chem.Mol", "Template molecule to do the embedding"),
		("neighbours", "list", "Idx of the atoms neighbouring the metal"),
		("name_input", "str", "Base name for the embedded molecules"),
		("maxsteps", "int", "Maximum number of optimization steps run with rdkit."),
		("log", "Logger", "[description]"),
	]
	item_fmt = "{} : {}\n    {}".format
	params_txt = "\n".join([item_fmt(*items) for items in parameters])
	f.__doc__ = f"{description}\nParameters\n----------\n{params_txt}\n"
	return f


def doc_returns(f):
	"""
	Decorator that adds the "Returns" section at the end of the
	docstrings of the decorated function.

	Parameters
	----------
	f : function
		function to decorate.

	Returns
	-------
	function
		returns the same function with the docstring modified.
	"""
	description = f.__doc__
	item_fmt = "{} : {}\n    {}".format
	outputs = [
		("mol_objects", "list", "Embedded molecules."),
		("name_returns", "list", "Names for the embedded molecules"),
		(
			"coord_maps",
			"list",
			"Mappings to the Idx of the atoms afected by the embedding",
		),
		("alg_maps", "list", "Mappings to the Idx of the core to do alignments"),
		("mol_templates", "list", "List of template molecules used"),
	]
	outs_txt = "\n".join([item_fmt(*items) for items in outputs])
	f.__doc__ = f"{description}\nReturns\n-------\n{outs_txt}\n"
	return f


# Embedding functions
@doc_returns
@doc_parameters
def two_embed(molecule, template, neighbours, name, maxsteps, log):
	"""
	Embedding function for linear geometries. Requires 'linear.sdf' template.
	"""
	template.GetAtomWithIdx(0).setAtomicNum(neighbours[0].GetAtomicNum())
	template.GetAtomWithIdx(1).setAtomicNum(neighbours[1].GetAtomicNum())
	template.GetAtomWithIdx(2).setAtomicNum(53)

	# assigning and embedding onto the core
	mol_obj, coord_map, alg_map, ci = template_embed_optimize(
		molecule, template, maxsteps, log
	)
	if ci >= 0:  # writing to mol_object file
		return [mol_obj], [name], [coord_map], [alg_map], [template]

	return [], [], [], [], []


@doc_returns
@doc_parameters
def three_embed(molecule, template, neighbours, name, maxsteps, log):
	"""
	Embedding function for trigonal planar geometry. Requires
	'trigonalplanar.sdf' template.
	"""
	template.GetAtomWithIdx(0).setAtomicNum(53)
	template.GetAtomWithIdx(1).setAtomicNum(neighbours[0].GetAtomicNum())
	template.GetAtomWithIdx(2).setAtomicNum(neighbours[1].GetAtomicNum())
	template.GetAtomWithIdx(3).setAtomicNum(neighbours[2].GetAtomicNum())

	# assigning and embedding onto the core
	mol_obj, coord_map, alg_map, conf_id = template_embed_optimize(
		molecule, template, maxsteps, log
	)
	if conf_id >= 0:  # writing to mol_object file
		return [mol_obj], [name], [coord_map], [alg_map], [template]

	return [], [], [], [], []


@doc_returns
@doc_parameters
def four_embed(molecule, template, neighbours, name, maxsteps, log):
	"""
	Embedding function for squareplanar geometry. Requires 'template-4-and-5.sdf'
	template. Attempts 3 embeddings.
	"""
	mol_objects = []
	name_return = []
	coord_maps = []
	alg_maps = []
	mol_templates = []

	# Remove F atoms from the template
	for atom in template.GetAtoms():
		if atom.GetSymbol() == "F":
			template = Chem.RWMol(template)
			template.RemoveAtom(atom.GetIdx())
	template = template.GetMol()

	# three cases for square planar
	atn0 = neighbours[0].GetAtomicNum()
	atn1 = neighbours[1].GetAtomicNum()
	atn2 = neighbours[2].GetAtomicNum()
	atn3 = neighbours[3].GetAtomicNum()
	replacements_list = [
		(atn0, atn1, atn2, atn3, 14),
		(atn0, atn2, atn3, atn1, 14),
		(atn0, atn3, atn1, atn2, 14),
	]

	for i, replacements in enumerate(replacements_list):

		# Create a copy of the mol object
		mol = Chem.Mol(template)

		# Assign atomic numbers to neighbour atoms
		for idx, atn in enumerate(replacements):
			mol.GetAtomWithIdx(idx).SetAtomicNum(atn)

		# embedding of the molecule onto the core
		mol_obj, coord_map, alg_map, conf_id = template_embed_optimize(
			molecule, mol, maxsteps, log
		)

		if conf_id >= 0:
			name_out = f"{name.split()[0]}_{i}"
			mol_objects.append(mol_obj)
			name_return.append(name_out)
			coord_maps.append(coord_map)
			alg_maps.append(alg_map)
			mol_templates.append(mol)

	return mol_objects, name_return, coord_maps, alg_maps, mol_templates


@doc_returns
@doc_parameters
def five_embed(molecule, mol_template, neighbours, name, maxsteps, log):
	"""
	Embedding function for squarepyramidal geometry. Requires
	'template-4-and-5.sdf' template. Attempts 15 embeddings.
	"""
	mol_objects = []
	name_return = []
	coord_maps = []
	alg_maps = []
	mol_templates = []
	counter = 0
	atomic_numbers = [mol_template.GetAtomWithIdx(i).GetAtomicNum() for i in neighbours]
	replacements = [
		[4, 0, 1, 2, 3],
		[4, 0, 2, 3, 1],
		[4, 0, 3, 1, 2],
		[3, 0, 1, 2, 4],
		[3, 0, 2, 4, 1],
		[3, 0, 4, 1, 2],
		[2, 0, 1, 4, 3],
		[2, 0, 4, 3, 1],
		[2, 0, 4, 1, 3],
		[1, 0, 4, 2, 3],
		[1, 0, 2, 3, 4],
		[1, 0, 3, 4, 2],
		[0, 4, 1, 2, 3],
		[0, 4, 2, 3, 1],
		[0, 4, 3, 1, 2],
	]
	for replacement in replacements:
		at0, at1, at2, at3, at4 = [atomic_numbers[r] for r in replacement]
		mol_template.GetAtomWithIdx(0).SetAtomicNum(at0)
		mol_template.GetAtomWithIdx(1).SetAtomicNum(at1)
		mol_template.GetAtomWithIdx(2).SetAtomicNum(at2)
		mol_template.GetAtomWithIdx(3).SetAtomicNum(at3)
		mol_template.GetAtomWithIdx(4).SetAtomicNum(at4)
		mol_template.GetAtomWithIdx(5).SetAtomicNum(14)
		mol_template.GetAtomWithIdx(5).SetFormalCharge(1)

		# assigning and embedding onto the core
		mol_obj, coord_map, alg_map, conf_id = template_embed_optimize(
			molecule, mol_template, maxsteps, log
		)
		if conf_id >= 0:
			name_out = f"{name}_{counter}"
			mol_objects.append(mol_obj)
			name_return.append(name_out)
			coord_maps.append(coord_map)
			alg_maps.append(alg_map)
			mol_templates.append(mol_template)
			counter += 1
	return mol_objects, name_return, coord_maps, alg_maps, mol_templates


# Main API of the module
def template_embed(
	molecule, name_input, log, complex_type, metal_idx, maxsteps, heavyonly, maxmatches
):
	"""
	Wrapper function to select automatically the appropiate embedding function
	depending on the number of neighbours of the metal center.

	Parameters
	----------
	molecule : [type]
		[description]
	name_input : [type]
		[description]
	log : [type]
		[description]
	complex_type : [type]
		[description]
	metal_idx : [type]
		[description]
	maxsteps : [type]
		[description]
	heavyonly : [type]
		[description]
	maxmatches : [type]
		[description]

	Returns
	-------
	mol_objects : list
		[description]
	name_return : list
		[description]
	coord_maps : list
		[description]
	alg_maps : list
		[description]
	mol_templates : list
		[description]
	"""
	embed_functions = dict()
	embed_functions[2] = two_embed
	embed_functions[3] = three_embed
	embed_functions[4] = four_embed
	embed_functions[5] = five_embed

	template = load_template(complex_type, log)

	# Generate the embeddings
	neighbours = calc_neighbours(molecule, metal_idx)
	embed = embed_functions[len(neighbours)]
	items = embed(molecule, template, neighbours, name_input, maxsteps, log)

	# Filter the results
	molecules = items[0]
	if len(molecules) > 1:
		ignored = []
		for i, mol in enumerate(molecules):
			has_big_rmsd = filter_template_mol(mol, molecules, heavyonly, maxmatches)
			if has_big_rmsd:
				ignored.append(i)
		items = [item for i, item in enumerate(items) if i not in ignored]
	return items
